from .base import Expr
from ..node import Node
from spartan.array import tile, distarray
import numpy as np

class NdArrayExpr(Expr):
  __metaclass__ = Node
  _members = ['_shape', 'dtype', 'tile_hint', 'combine_fn', 'reduce_fn']

  def __str__(self):
    return 'array(%s, %s)' % (self.shape, self.dtype)
  
  def visit(self, visitor):
    return NdArrayExpr(_shape=visitor.visit(self.shape),
                       dtype=visitor.visit(self.dtype),
                       tile_hint=self.tile_hint,
                       combine_fn=self.combine_fn,
                       reduce_fn=self.reduce_fn)
  
  def dependencies(self):
    return {}
  
  def compute_shape(self):
    return self._shape
 
  def evaluate(self, ctx, deps):
    shape = self._shape
    dtype = self.dtype
    tile_hint = self.tile_hint
    
    return distarray.create(shape, dtype,
                            combiner=self.combine_fn,
                            reducer=self.reduce_fn,
                            tile_hint=tile_hint)


def ndarray(shape, 
            dtype=np.float, 
            tile_hint=None,
            combine_fn=None, 
            reduce_fn=None):
  '''
  Lazily create a new distributed array.
  :param shape:
  :param dtype:
  :param tile_hint:
  '''
  return NdArrayExpr(_shape = shape,
                     dtype = dtype,
                     tile_hint = tile_hint,
                     combine_fn = combine_fn,
                     reduce_fn = reduce_fn) 



from .base import Expr
from spartan import util
from spartan.array import extent, tile, distarray
from spartan.expr.base import make_primitive
from spartan.node import Node


def _reduce_mapper(ex, input, reducer, axis, output, fn_kw):
  #util.log_info('Reduce: %s %s %s %s %s', reducer, ex, tile, axis, fn_kw)
  tile = input.fetch(ex)
  reduced = reducer(ex, tile, axis, **fn_kw)
  dst_extent = extent.index_for_reduction(ex, axis)
  #util.log_info('Update: %s %s', dst_extent, reduced)
  output.update(dst_extent, reduced)
  return []

class ReduceExpr(Expr):
  __metaclass__ = Node
  _members = ['array', 'axis', 'dtype_fn', 'reduce_fn', 'combine_fn', 'fn_kw']
  
  def evaluate(self, ctx, deps):
    input_array = deps['array']
    dtype = deps['dtype_fn'](input_array)
    axis = deps['axis']
    reducer = deps['reduce_fn']
    fn_kw = deps['fn_kw']
    
    util.log_info('Reducing %s over axis %s', input_array.shape, axis)
    
    shape = extent.shape_for_reduction(input_array.shape, axis)
    tile_accum = deps['combine_fn']
    
    output_array = distarray.create(shape, dtype,
                                    combiner=tile_accum, 
                                    reducer=tile_accum)
    
    
    util.log_info('Reducing into array %s', output_array)
    input_array.foreach(_reduce_mapper, kw={
      'input' : input_array,
      'reducer' : reducer,
      'axis' : axis,
      'output' : output_array,
      'fn_kw' : fn_kw})
    
    return output_array

 
def reduce(v, axis, dtype_fn, local_reduce_fn, combine_fn, fn_kw=None):
  '''
  Reduce ``v`` over axis ``axis``.
  
  The resulting array should have a datatype given by ``dtype_fn(input).``
  
  For each tile of the input ``local_reduce_fn`` is called.
  The output is combined using ``combine_fn``.
   
  :param v: `Expr`
  :param axis: int or None
  :param dtype_fn: Callable: fn(array) -> `numpy.dtype`
  :param local_reduce_fn: Callable: fn(extent, data, axis)
  :param combine_fn: Callable: fn(old_v, update_v) -> new_v 
  '''

  if fn_kw is None: fn_kw = {}
  return ReduceExpr(array=v, axis=axis, 
                     dtype_fn=dtype_fn, 
                     reduce_fn=local_reduce_fn, 
                     combine_fn=combine_fn,
                     fn_kw=fn_kw)



from spartan import blob_ctx
from spartan.array import distarray, tile
from spartan.node import Node
from spartan.util import iterable

from .base import Expr, lazify


def shuffle(v, fn, tile_hint=None, target=None, kw=None):
  '''
  Evaluate ``fn`` over each extent of the input.
  
  ``fn`` should take arguments: ``(v, extent, **kw)``
  
  :param v: `Expr` to map over.
  :param fn: Callable with form ``(v, extent, **kw)``
  '''
  if kw is None: kw = {}
  
  kw = lazify(kw)
  v = lazify(v)
  #util.log_info('%s', kw)
  
  assert not iterable(v)
  
  return ShuffleExpr(array=v,
                     map_fn=fn,
                     tile_hint=tile_hint,
                     target=target,
                     fn_kw=kw)


def _target_mapper(ex, map_fn=None, inputs=None, target=None, fn_kw=None):
  result = map_fn(inputs, ex, **fn_kw)
  if result is not None:
    for ex, v in result:
      target.update(ex, v)
  return []
        
def _notarget_mapper(ex, array=None, map_fn=None, inputs=None, fn_kw=None):
  #util.log_info('MapExtents: %s', map_fn)
  ctx = blob_ctx.get()
  result = []
  map_result = map_fn(inputs, ex, **fn_kw)
  if map_result is not None:
    for ex, v in map_result:
      result.append((ex, v))
  return result

class ShuffleExpr(Expr):
  __metaclass__ = Node
  _members = ['array', 'map_fn', 'target', 'tile_hint', 'fn_kw']

  def __str__(self):
    return 'shuffle[%d](%s, %s)' % (id(self), self.map_fn, self.array)

    
  def evaluate(self, ctx, deps):
    v = deps['array']
    fn_kw = deps['fn_kw']
    target = deps['target']
    
    map_fn = self.map_fn

    if target is not None:
      v.foreach(_target_mapper,
                kw = dict(map_fn=map_fn, inputs=v, target=target, fn_kw=fn_kw))
      return target
    else:
      return v.map_to_array(mapper_fn = _notarget_mapper,
                            kw = dict(inputs=v, map_fn=map_fn, fn_kw=fn_kw))



from . import util, rpc, core
import threading
from .util import Assert

MASTER_ID = 65536
ID_COUNTER = iter(xrange(10000000))

class BlobCtx(object):
  def __init__(self, worker_id, workers, local_worker=None):
    assert isinstance(workers, dict)
    assert isinstance(worker_id, int)

    self.num_workers = len(workers)
    self.worker_id = worker_id
    self.workers = workers
    self.id_map = {}
    self.local_worker = local_worker
    self.active = True
    self._deferred = []

    #util.log_info('New blob ctx.  Worker=%s', self.worker_id)

  def defer(self, closure):
    self._deferred.append(closure)

  def _send(self, id, method, req, wait=True):
    if self.active == False:
      util.log_debug('Ctx disabled.')
      return None

    # process any deferred operations now:
    if self._deferred:
      fns = self._deferred
      self._deferred = []
      for fn in fns:
        #util.log_info('Running deferred operation %s', fn)
        fn()

    #util.log_info('%s %s', id, method)
    worker_id = self._lookup(id)

    # short-circuit for requests to the local worker
    if worker_id == self.worker_id:
      pending_req = rpc.Future(None, -1)
      getattr(self.local_worker, method)(req, pending_req)
    else:
      pending_req =  getattr(self.workers[worker_id], method)(req)

    if wait:
      return pending_req.wait()

    return pending_req

  def _send_all(self, method, req, wait=True):
    if self.active == False:
      util.log_debug('Ctx disabled.')
      return None

    futures = rpc.forall(self.workers.itervalues(), method, req)
    if wait:
      return futures.wait()
    return futures

  def _lookup(self, blob_id):
    worker_id = blob_id.worker
    return worker_id
    #if not blob_id in self.id_map:
    #futures = [(id, w.lookup(blob_id)) for id, w in self.workers.iteritems()]
    #for idx, f in enumerate(futures):
    #  if f.wait() is not None:
    #    self.id_map[blob_id] = idx
    #return self.id_map[blob_id]

  def destroy_all(self, blob_ids):
    if self.worker_id != MASTER_ID: return
    #util.log_info('Destroy: %s', blob_ids)
    req = core.DestroyReq(ids=blob_ids)

    # fire and forget...?
    self._send_all('destroy', req, wait=False)

  def destroy(self, blob_id):
    return self.destroy_all([blob_id])

  def get(self, blob_id, selector, callback=None):
    Assert.isinstance(blob_id, core.BlobId)
    req = core.GetReq(id=blob_id, selector=selector)
    if callback is None:
      return self._send(blob_id, 'get', req).data
    else:
      future = self._send(blob_id, 'get', req, wait=False)
      return future.on_finished(callback)


  def update(self, blob_id, data, reducer, wait=True):
    req = core.UpdateReq(id=blob_id, data=data, reducer=reducer)
    return self._send(blob_id, 'update', req, wait=wait)

  def create_local(self):
    assert self.worker_id != MASTER_ID
    return core.BlobId(worker=self.worker_id, id=ID_COUNTER.next())

  def create(self, data, hint=None):
    assert self.worker_id >= 0, self.worker_id

    if self.worker_id >= MASTER_ID:
      if hint is None:
        worker_id = ID_COUNTER.next() % len(self.workers)
      else:
        worker_id = hint % len(self.workers)
      id = -1
    else:
      worker_id = self.worker_id
      id = ID_COUNTER.next()

    blob_id = core.BlobId(worker=worker_id, id=id)

    #util.log_info('%s %s %s %s', new_id, worker_id, data.shape, ''.join(traceback.format_stack()))

    req = core.CreateReq(blob_id=blob_id, data=data)
    return self._send(blob_id, 'create', req, wait=False)

  def map(self, blob_ids, mapper_fn, reduce_fn, kw):
    if self.active == False:
      util.log_debug('Ctx disabled.')
      return None

    req = core.KernelReq(blobs=blob_ids,
                    mapper_fn=mapper_fn,
                    reduce_fn=reduce_fn,
                    kw=kw)

    #util.log_info('%s', req)

    futures = rpc.forall(self.workers.itervalues(), 'run_kernel', req).wait()
    result = {}
    for f in futures:
      for blob_id, v in f.iteritems():
        result[blob_id] = v
    return result


_ctx = threading.local()

def get():
  return _ctx.val

def set(ctx):
  _ctx.val = ctx




#!/usr/bin/env python

'''
Evalution for `Expr` nodes.

`evaluate` evaluates a nodes dependencies, caching
results, then evaluates nodes themselves.
'''

from spartan import util
from spartan.util import Assert

def _evaluate(ctx, prim):
  #util.log_info('Evaluting deps for %s', prim)
  deps = {}
  for k, vs in prim.dependencies().iteritems():
    deps[k] = evaluate(ctx, vs)
      
  #util.log_info('Evaluting %s', prim.typename())
  return prim.evaluate(ctx, deps)
  #return util.timeit(lambda: prim.evaluate(ctx, deps), 'eval: %s' % prim)

def evaluate(ctx, prim):
  '''
  Evaluate an `Expr`.  
 
  Dependencies are evaluated prior to evaluating ``prim``.
   
  Dependencies may be either a list, a dictionary or a single
  value of type `Expr`.  For convenience, we allow specifying 
  non-Expr dependencies; these are left unaltered.
  
  :param ctx:
  :param prim: `Expr`
  '''
  from .base import Expr
  
  #util.log_info('%s', type(prim))
  if isinstance(prim, dict):
    return dict([(k, evaluate(ctx, v)) for (k, v) in prim.iteritems()])
  elif isinstance(prim, tuple):
    return tuple([evaluate(ctx, v) for v in prim])
  elif isinstance(prim, list):
    return [evaluate(ctx, v) for v in prim]
  elif not isinstance(prim, Expr):
    return prim
  
  if prim._cached_value is None:
    #util.log_info('Evaluating: %s', prim)
    prim._cached_value = _evaluate(ctx, prim)
  
  return prim._cached_value




#!/usr/bin/env python

import collections
from spartan import util
from spartan.util import Assert
import numpy as np

class TileExtent(object):
  '''A rectangular tile of a distributed array.
  
  These correspond (roughly) to a `slice` taken from an array
  (without any step component).
  
  Arrays are indexed from the upper-left; for an array of shape
  (sx, sy, sz): (0,0...) is the upper-left corner of an array, 
  and (sx,sy,sz...) the lower-right.
  
  Extents are represented by an upper-left corner (inclusive) and
  a lower right corner (exclusive): [ul, lr).  In addition, they
  carry the shape of the array they are a part of; this is used to
  compute global position information.
  '''

  @property
  def size(self):
    return np.prod(self.shape)
  
  @property
  def shape(self):
    result = np.asarray(self.lr) - np.asarray(self.ul)
    result[result == 0] = 1
    #util.log_info('Shape: %s', result)
    return tuple(result)
  
  @property
  def ndim(self):
    return len(self.lr)

  def __reduce__(self):
    return create, (self.ul, self.lr, self.array_shape)
  
  def to_slice(self):
    return tuple([slice(ul, lr, None) for ul, lr in zip(self.ul, self.lr)])
  
  def __repr__(self):
    return 'extent(' + ','.join('%s:%s' % (a, b) for a, b in zip(self.ul, self.lr)) + ')'

  
  def __getitem__(self, idx):
    return create((self.ul[idx],),
                  (self.lr[idx],),
                  (self.array_shape[idx],))

  def __hash__(self):
    return hash(self.ul)
    #return hash(self.ul[-2:])
    #return ravelled_pos(self.ul, self.array_shape)
    
  def __eq__(self, other):
    return self.ul == other.ul and self.lr == other.lr

  def ravelled_pos(self):
    return ravelled_pos(self.ul, self.array_shape)
  
  def to_global(self, idx, axis):
    '''Convert ``idx`` from a local offset in this tile to a global offset.'''
    if axis is not None:
      return idx + self.ul[axis]

    local_idx = unravelled_pos(idx, self.shape)
    return ravelled_pos(np.asarray(self.ul) + local_idx, self.array_shape)

  def add_dim(self):
    return create(self.ul + (0,), 
                  self.lr + (0,), 
                  self.array_shape + (1,))

  def clone(self):
    return create(self.ul, self.lr, self.array_shape)
 
#import traceback
counts = collections.defaultdict(int)

def create(ul, lr, array_shape):
  '''
  Create a new extent with the given coordinates and array shape.
  
  :param ul: `tuple`: 
  :param lr:
  :param array_shape:
  '''
  #stack = ''.join(traceback.format_stack())
  #counts[stack] += 1
  ex = TileExtent()
  ex.ul = tuple(ul)
  ex.lr = tuple(lr)

  #assert np.all(np.array(ex.lr) >= np.array(ex.ul)),\
  #  'Negative extent size: (%s, %s)' % (ul, lr)
  
  if array_shape is not None:
    ex.array_shape = tuple(array_shape)
    #assert np.all(np.array(ex.lr) <= np.array(array_shape)),\
    #  'Extent lr (%s) falls outside of the array(%s)' % (lr, array_shape)
  else:
    ex.array_shape = None
  
  return ex

def from_shape(shp):
  return create(tuple([0] * len(shp)), tuple(v for v in shp), shp)

 
def unravelled_pos(idx, array_shape): 
  '''
  Unravel ``idx`` into an index into an array of shape ``array_shape``.
  :param idx: `int`
  :param array_shape: `tuple`
  :rtype: `tuple` indexing into ``array_shape``
  '''
  
  unravelled = []
  for dim in reversed(array_shape):
    unravelled.append(idx % dim)
    idx /= dim
  
  return tuple(reversed(unravelled))
    
def ravelled_pos(idx, array_shape):
  rpos = 0
  mul = 1
  
  for i in range(len(array_shape) - 1, -1, -1):
    rpos += mul * idx[i]
    mul *= array_shape[i]
  
  return rpos

def find_overlapping(extents, region):
  '''
  Return the extents that overlap with ``region``.   
  
  :param extents: List of extents to search over.
  :param region: `Extent` to match.
  '''
  for ex in extents:
    overlap = intersection(ex, region)
    if overlap is not None:
      yield (ex, overlap)
      

def compute_slice(base, idx):
  '''Return a new ``TileExtent`` representing ``base[idx]``
  
  :param base: `TileExtent`
  :param idx: int, slice, or tuple(slice,...)
  '''
  assert not np.isscalar(idx), idx
  if not isinstance(idx, tuple):
    idx = (idx,)
    
  ul = []
  lr = []
  array_shape = base.array_shape
  
  for i in range(len(base.ul)):
    if i >= len(idx):
      ul.append(base.ul[i])
      lr.append(base.lr[i])
    else:
      start, stop, step = idx[i].indices(base.shape[i])
      ul.append(base.ul[i] + start)
      lr.append(base.ul[i] + stop)
  
  return create(ul, lr, array_shape)


def offset_from(base, other):
  '''
  :param base: `TileExtent` to use as basis
  :param other: `TileExtent` into the same array.
  :rtype: A new extent using this extent as a basis, instead of (0,0,0...) 
  '''
  assert np.all(other.ul >= base.ul), (other, base)
  assert np.all(other.lr <= base.lr), (other, base)
  return create(tuple(np.array(other.ul) - np.array(base.ul)),
                tuple(np.array(other.lr) - np.array(base.ul)),
                other.array_shape)



def offset_slice(base, other):
  '''
  :param base: `TileExtent` to use as basis
  :param other: `TileExtent` into the same array.
  :rtype: A slice representing the local offsets of ``other`` into this tile.
  '''
  return offset_from(base, other).to_slice()
  

def from_slice(idx, shape):
  '''
  Construct a `TileExtent` from a slice or tuple of slices.
  
  :param idx: int, slice, or tuple(slice...)
  :param shape: shape of the input array
  :rtype: `TileExtent` corresponding to ``idx``.
  '''
  if not isinstance(idx, tuple):
    idx = (idx,)
  
  if len(idx) < len(shape):
    idx = tuple(list(idx) + [slice(None, None, None) 
                             for _ in range(len(shape) - len(idx))])
    
  ul = []
  lr = []
 
  for i in range(len(shape)):
    dim = shape[i]
    slc = idx[i]
    
    if np.isscalar(slc):
      slc = int(slc)
      slc = slice(slc, slc + 1, None)
    
    if slc.start > 0: assert slc.start <= dim
    if slc.stop > 0: assert slc.stop <= dim
    
    indices = slc.indices(dim)
    ul.append(indices[0])
    lr.append(indices[1])
    
  return create(tuple(ul), tuple(lr), shape)


def intersection(a, b):
  '''
  :rtype: The intersection of the 2 extents as a `TileExtent`, 
          or None if the intersection is empty.  
  '''
  for i in range(len(a.lr)):
    if b.lr[i] < a.ul[i]: return None
    if a.lr[i] < b.ul[i]: return None
    
  Assert.eq(a.array_shape, b.array_shape)
  
  return create(np.maximum(b.ul, a.ul),
                np.minimum(b.lr, a.lr),
                a.array_shape)


def shape_for_reduction(input_shape, axis):
  '''
  Return the shape for the result of applying a reduction along ``axis`` to 
  an input of shape ``input_shape``.
  :param input_shape:
  :param axis:
  '''
  if axis == None: return ()
  input_shape = list(input_shape)
  del input_shape[axis]
  return input_shape


def shapes_match(offset, data):
  '''
  Return true if the shape of ``data`` matches the extent ``offset``. 
  :param offset:
  :param data:
  '''
  return np.all(offset.shape == data.shape)

def drop_axis(ex, axis):
  if axis is None: return create((), (), ())
  if axis < 0: axis = len(ex.ul) + axis
  
  ul = list(ex.ul)
  lr = list(ex.lr)
  shape = list(ex.array_shape)
  del ul[axis]
  del lr[axis]
  del shape[axis]
  return create(ul, lr, shape)
 
def index_for_reduction(index, axis):
  return drop_axis(index, axis)
        
def find_shape(extents):
  '''
  Given a list of extents, return the shape of the array
  necessary to fit all of them.
  :param extents:
  '''
  #util.log_info('Finding shape... %s', extents)
  shape = np.max([ex.lr for ex in extents], axis=0)
  shape[shape == 0] = 1
  return tuple(shape)


def is_complete(shape, slices):
  '''
  Returns true if ``slices`` is a complete covering of shape; that is:

  ::

    array[slices] == array

  :param shape: tuple of int
  :param slices: list/tuple of `slice` objects
  :rtype: boolean
  '''
  if len(shape) != len(slices):
    return False

  for dim,slice in zip(shape, slices):
    if slice.start > 0: return False
    if slice.stop < dim: return False
  return True



#!/usr/bin/env python

'''
Optimizations over an expression graph.
'''

import numpy as np
from spartan.config import FLAGS, BoolFlag
from spartan.array import extent
from spartan.expr.reduce import ReduceExpr
from spartan.node import Node
from spartan.util import Assert, iterable

from .. import util
from .base import Expr, LazyVal, LazyList, AsArray
from .shuffle import ShuffleExpr
from .map import MapExpr
from .ndarray import NdArrayExpr
from spartan.array.distarray import broadcast

try:
  import numexpr
except:
  numexpr = None
  
# Local versions of expression graph operations.
# These are used to build up fusion operations.

class LocalOp(Expr):
  __metaclass__ = Node
  _cached_value = None
  def evaluate(self, ctx):
    deps = {}
    for k, v in self.dependencies():
      if not iterable(v): deps[k] = v.evaluate()
      else: deps[k] = [sv.evaluate() for sv in v]
    
    self._cached_value = self._evaluate(ctx, deps)
    return self._cached_value
      

class LVal(LocalOp):
  '''
  A local value (a constant, or an input argument).
  '''
  _members = ['source', 'id']
  
  def dependencies(self): return {}
   
  def _evaluate(self, ctx, deps):
    return ctx[self.source][self.id]


class LMap(LocalOp):
  _members = ['fn', 'inputs', 'kw']
  
  def _evaluate(self, ctx, deps):
    return deps['fn'](deps['inputs'], **deps['kw'])
    
class LReduce(LocalOp):
  _members = ['fn', 'inputs', 'kw']
  
  def _evaluate(self, ctx, deps):
    return deps['fn'](deps['inputs'], ctx['extent'], **deps['kw'])


class OptimizePass(object):
  def __init__(self):
    self.visited = {}
  
  def visit(self, op):
    if not isinstance(op, Expr):
      return op
    
    if op in self.visited:
      return self.visited[op]
    
    #assert not op in self.visited, 'Infinite recursion during optimization %s' % op
    #self.visited.add(op)
    
    #util.log_info('VISIT %s: %s', op.typename(), hash(op))
    if hasattr(self, 'visit_%s' % op.typename()):
      self.visited[op] = getattr(self, 'visit_%s' % op.typename())(op)
    else:
      self.visited[op] = op.visit(self)
      
    return self.visited[op]


def _dag_eval(inputs, dag, kw_inputs):
  '''
  Function for evaluated folded operations.
  
  Takes 
  :param inputs:
  :param dag:
  '''
  ctx = { 'kw' : kw_inputs,
          'inputs' : inputs }
  return dag.evaluate(ctx, {})
    

def _fold_mapper(*inputs,  **kw):
  '''Helper mapper function for folding.
  
  Runs each callable in `fns` on a number of the input tiles.
  
  :param inputs: Input values for folded mappers
  :param fns: A list of dictionaries containing { 'fn', 'fn_kw', 'range' }
  :param map_fn:
  :param map_kw:
  '''
  fns = kw['fns']
  map_fn = kw['map_fn']
  map_kw = kw['map_kw']
  results = []
  for fn_info in fns:
    st, ed = fn_info['range']
    fn = fn_info['fn']
    kw = fn_info['fn_kw']
    fn_inputs = inputs[st:ed]
    result = fn(*fn_inputs, **kw)
    assert isinstance(result, np.ndarray) or np.isscalar(result), (type(result), fn)
    results.append(result)
  
  # util.log_info('%s %s %s', map_fn, results, map_kw)
  if 'numpy_expr' in map_kw: del map_kw['numpy_expr']
  return map_fn(*results, **map_kw)
    

def _take_first(*lst):
#  util.log_info('Take first: %s --> %s', lst, lst[0])
  return lst[0]

def map_like(v):
  return isinstance(v, (MapExpr, ShuffleExpr, NdArrayExpr, LazyVal, AsArray))


class MapMapFusion(OptimizePass):
  '''Fold sequences of Map operations together.
  
  map(f, map(g, map(h, x))) -> map(f . g . h, x)
  '''
  
  name = 'map_fusion'
   
  def visit_MapExpr(self, op):
    #util.log_info('Map tiles: %s', op.children)
    Assert.iterable(op.children)
    #util.log_info('VISIT %d', id(op))
    #util.log_info('VISIT %s', op.children)
    map_children = self.visit(op.children)
    all_maps = np.all([map_like(v) for v in map_children])
    
    #util.log_info('Folding: %s %s', [type(v) for v in map_children], all_maps)
    if not all_maps:
      return op.visit(self)
    
    children = []
    fns = []
    for v in map_children:
      op_st = len(children)
      
      if isinstance(v, MapExpr):
        op_in = self.visit(v.children)
        children.extend(op_in)
        map_fn = v.map_fn
        fn_kw = v.fn_kw
      else:
        # evaluate these operations directly and use the result; 
        # we can use the input of these operations, but can't
        # avoid creating a new array.
        children.append(v)
        map_fn = _take_first
        fn_kw = {}
      
      op_ed = len(children)
      fns.append({ 'fn' : map_fn, 'fn_kw' : fn_kw, 'range' : (op_st, op_ed) }) 
    
    map_fn = op.map_fn
    map_kw = op.fn_kw
    # util.log_info('Map function: %s, kw: %s', map_fn, map_kw)
    
    # util.log_info('Created fold mapper with %d children', len(children))
    return MapExpr(children=LazyList(vals=children),
                   map_fn=_fold_mapper,
                   fn_kw={ 'fns' : fns,
                           'map_fn' : map_fn,
                           'map_kw' : map_kw })

def _folded_reduce(ex, tile, axis,
                   map_inputs, map_fn, map_kw,
                   reduce_fn, reduce_kw,):
  map_inputs = broadcast(map_inputs)
  local_values = [v.fetch(ex) for v in map_inputs]
  map_output = map_fn(*local_values, **map_kw)
  return reduce_fn(ex, map_output, axis, **reduce_kw)

class ReduceMapFusion(OptimizePass):
  name = 'reduce_fusion'
  
  def visit_ReduceExpr(self, op):
    array = self.visit(op.array)
    
    if isinstance(array, MapExpr):
      children = array.children
      Assert.isinstance(children, LazyList)
      return ReduceExpr(array=children[0],
                        axis=op.axis,
                        dtype_fn=op.dtype_fn,
                        reduce_fn=_folded_reduce,
                        combine_fn=op.combine_fn,
                        fn_kw={'map_kw' : array.fn_kw,
                               'map_fn' : array.map_fn,
                               'reduce_fn' : op.reduce_fn,
                               'reduce_kw' : op.fn_kw,
                               'map_inputs' : children })
    else:
      return op.visit(self)



def _numexpr_mapper(inputs, var_map=None, numpy_expr=None):
  gdict = {}
  for k, v in var_map.iteritems():
    gdict[k] = inputs[v]
    
  numexpr.ncores = 1 
  result = numexpr.evaluate(numpy_expr, global_dict=gdict)
  return result


_COUNTER = iter(xrange(1000000))
def new_var():
  return 'input_%d' % _COUNTER.next()
    

class NumexprFusionPass(OptimizePass):
  '''Fold binary operations compatible with numexpr into a single numexpr operator.'''
  name = 'numexpr_fusion'
  
  def visit_MapExpr(self, op):
    map_children = [self.visit(v) for v in op.children]
    all_maps = np.all([map_like(v) for v in map_children])
   
    if not (all_maps and op.numpy_expr is not None):
      return op.visit(self)
    
    a, b = map_children
    operation = op.numpy_expr
   
    # mapping from variable name to input index 
    var_map = {}
    
    # inputs to the expression
    inputs = []
    expr = []
    
    def _add_expr(child):
      # fold expression from the a mapper into this one.
      if isinstance(child, MapExpr) and child.map_fn == _numexpr_mapper:
        for k, v in child.fn_kw['var_map'].iteritems():
          var_map[k] = len(inputs)
          inputs.append(child.children[v])
        expr.extend(['(' + child.numpy_expr + ')'])
      else:
        v = new_var()
        var_map[v] = len(inputs)
        inputs.append(child)
        expr.append(v)
    
    _add_expr(a)
    expr.append(operation)
    _add_expr(b)
    
    expr = ' '.join(expr)
    
    return MapExpr(children=inputs,
                   map_fn=_numexpr_mapper,
                   numpy_expr = expr,
                   fn_kw={ 'numpy_expr' : expr, 'var_map' : var_map, })


def apply_pass(klass, dag):
  if not getattr(FLAGS, 'opt_' + klass.name):
    util.log_debug('Pass %s disabled', klass.name)
    return dag
  
  p = klass()
  return p.visit(dag)

passes = []

def optimize(dag):
  if not FLAGS.optimization:
    util.log_info('Optimizations disabled')
    return dag
  
  for p in passes:
    dag = apply_pass(p, dag)
  
  return dag

def add_optimization(klass, default):
  passes.append(klass)
  
  flagname = 'opt_' + klass.name
  #setattr(Flags, flagname, add_bool_flag(flagname, default=default))
  FLAGS.add(BoolFlag(flagname, default=default, help='Enable %s optimization' % klass.__name__))

add_optimization(MapMapFusion, True)
add_optimization(ReduceMapFusion, True)
add_optimization(NumexprFusionPass, False)

FLAGS.add(BoolFlag('optimization', default=True))



'''
Simple RPC library.

The :class:`.Client` and :class:`.Server` classes here work with
sockets which should implement the :class:`.Socket` interface.
'''
from cPickle import PickleError
import collections
import weakref
import sys
import threading
import time
import traceback
import types
import cStringIO
import cPickle

from .. import cloudpickle, util, core
from ..node import Node

CLIENT_PENDING = weakref.WeakKeyDictionary()
SERVER_PENDING = weakref.WeakKeyDictionary()

NO_RESULT = object()

DEFAULT_TIMEOUT = 100

def set_default_timeout(seconds):
  global DEFAULT_TIMEOUT
  DEFAULT_TIMEOUT = seconds
  util.log_info('Set default timeout to %s seconds.', DEFAULT_TIMEOUT)


class RPCException(object):
  __metaclass__ = Node
  _members = ['py_exc']

class PickledData(object):
  '''
  Helper class: indicates that this message has already been pickled,
  and should be sent as is, rather than being re-pickled.
  '''
  __metaclass__ = Node
  _members = ['data']

class SocketBase(object):
  def send(self, blob): assert False
  def recv(self): assert False
  def flush(self): assert False
  def close(self): assert False

  def register_handler(self, handler):
    'A handler() is called in response to read requests.'
    self._handler = handler

  # client
  def connect(self): assert False

  # server
  def bind(self): assert False


def capture_exception(exc_info=None):
  if exc_info is None:
    exc_info = sys.exc_info()
  tb = traceback.format_exception(*exc_info)
  return RPCException(py_exc=''.join(tb).replace('\n', '\n:: '))


class Group(tuple):
  pass

def serialize_to(obj, writer):
  #serialization.write(obj, writer)
  try:
    pickled = cPickle.dumps(obj, -1)
    writer.write(pickled)
  except (PickleError, TypeError):
    #util.log_warn('CPICKLE failed: %s (%s)', sys.exc_info(), obj)
    writer.write(cloudpickle.dumps(obj, -1))

def serialize(obj):
  #x = cStringIO.StringIO()
  #serialization.write(obj, x)
  #return x.getvalue()
  try:
    return cPickle.dumps(obj, -1)
  except (PickleError, TypeError):
    return cloudpickle.dumps(obj, -1)

def read(f):
  return cPickle.load(f)

class PendingRequest(object):
  '''An outstanding RPC request on the server.

  Call done(result) when finished to send result back to client.
  '''
  def __init__(self, socket, rpc_id):
    self.socket = socket
    self.rpc_id = rpc_id
    self.created = time.time()
    self.finished = False
    self.result = NO_RESULT

    SERVER_PENDING[self] = 1

  def wait(self):
    while self.result is NO_RESULT:
      time.sleep(0.1)
    return self.result

  def exception(self):
    self.done(capture_exception())

  def done(self, result=None):
    # util.log_info('RPC finished in %.3f seconds' % (time.time() - self.created))
    self.finished = True
    self.result = result

    assert self.socket is not None
    header = { 'rpc_id' : self.rpc_id }
    #util.log_info('Finished %s, %s', self.socket.addr, self.rpc_id)
    w = cStringIO.StringIO()
    cPickle.dump(header, w, -1)
    serialize_to(result, w)
    self.socket.send(w.getvalue())

  def __del__(self):
    if not self.finished:
      util.log_error('PendingRequest.done() not called before destruction (likely due to an exception.)')
      self.done(result=RPCException(py_exc='done() not called on request.'))


class RemoteException(Exception):
  '''Wrap a uncaught remote exception.'''
  def __init__(self, tb):
    self._tb = tb

  def __repr__(self):
    return 'RemoteException:\n' + self._tb

  def __str__(self):
    return repr(self)


class Future(object):
  def __init__(self, addr, rpc_id):
    self.addr = addr
    self.rpc_id = rpc_id
    self.have_result = False
    self.result = None
    self.finished_fn = None
    self._cv = threading.Condition()
    self._start = time.time()
    self._deadline = time.time() + DEFAULT_TIMEOUT

    CLIENT_PENDING[self] = 1

  def done(self, result=None):
    #util.log_info('Result... %s %s', self.addr, self.rpc_id)
    self._cv.acquire()
    self.have_result = True

    if self.finished_fn is not None:
      self.result = self.finished_fn(result)
    else:
      self.result = result

    self._cv.notify()
    self._cv.release()

  def timed_out(self):
    return self._deadline < time.time()

  def wait(self):
    self._cv.acquire()
    while not self.have_result and not self.timed_out():
      # use a timeout so that ctrl-c works.
      self._cv.wait(timeout=1)
      if time.time() - self._start > 2:
        util.log_info('Waiting... %s %s', self.addr, self.rpc_id)
    self._cv.release()

#    util.log_info('Result from %s in %f seconds.', self.addr, time.time() - self._start)

    if not self.have_result and self.timed_out():
      util.log_info('timed out!')
      raise Exception('Timed out on remote call (%s %s)', self.addr(), self.rpc_id)

    if isinstance(self.result, RPCException):
      raise RemoteException(self.result.py_exc)
    return self.result

  def on_finished(self, fn):
    return FnFuture(self, fn)


class FnFuture(object):
  '''Chain ``fn`` to the given future.

  ``self.wait()`` return ``fn(future.wait())``.
  '''
  def __init__(self, future, fn):
    self.future = future
    self.fn = fn
    self.result = None

  def wait(self):
    return self.fn(self.future.wait())


class FutureGroup(list):
  def wait(self):
    return [f.wait() for f in self]

def wait_for_all(futures):
  return [f.wait() for f in futures]


class Server(object):
  def __init__(self, socket):
    self._socket = socket
    self._socket.register_handler(self.handle_read)
    self._methods = {}
    self._timers = collections.defaultdict(util.Timer)
    self._running = False
    self.register_method('diediedie', self._diediedie)


  def timings(self):
    return '\n'.join(['%s: %f' % (m, self._timers[m].get())
                      for m in self._methods.keys()])

  def _diediedie(self, handle, req):
    handle.done(None)
    self._socket.flush()
    self.shutdown()

  @property
  def addr(self):
    return self._socket.addr

  def serve(self):
    self.serve_nonblock()
    while self._running:
      time.sleep(1)

  def serve_nonblock(self):
#    util.log_info('Running.')
    self._running = True
    self._socket.bind()

  def register_object(self, obj):
    for name in dir(obj):
      if name.startswith('__'): continue
      fn = getattr(obj, name)
      if isinstance(fn, types.MethodType):
        self.register_method(name, fn)
  
  def register_method(self, name, fn):
    self._methods[name] = fn

  def handle_read(self, socket):
    #util.log_info('Reading...')

    data = socket.recv()
    reader = cStringIO.StringIO(data)
    header = cPickle.load(reader)

    #util.log_info('Call[%s] %s %s', header['method'], self._socket.addr, header['rpc_id'])
    handle = PendingRequest(socket, header['rpc_id'])
    name = header['method']

    try:
      fn = self._methods[name]
    except KeyError:
      handle.exception()
      return

    try:
      req = read(reader)
      result = fn(req, handle)
      assert result is None, 'non-None result from RPC handler (use handle.done())'
    except:
      util.log_info('Caught exception in handler.', exc_info=1)
      handle.exception()

  def shutdown(self):
    self._running = 0
    self._socket.close()
    del self._socket


class ProxyMethod(object):
  def __init__(self, client, method):
    self.client = client
    self.method = method

  def __call__(self, request=None):
#    if len(serialized) > 800000:
#      util.log_info('%s::\n %s; \n\n\n %s', self.method, ''.join(traceback.format_stack()), request)
    return self.client.send(self.method, request)

class Client(object):
  def __init__(self, socket):
    self._socket = socket
    self._socket.register_handler(self.handle_read)
    self._socket.connect()
    self._futures = {}
    self._lock = threading.Lock()
    self._rpc_id = xrange(10000000).__iter__()

  def __reduce__(self, *args, **kwargs):
    raise cPickle.PickleError('Not pickleable.')

  def send(self, method, request):
    with self._lock:
      rpc_id = self._rpc_id.next()
      header = { 'method' : method, 'rpc_id' : rpc_id }

      w = cStringIO.StringIO()
      cPickle.dump(header, w, -1)
      if isinstance(request, PickledData):
        w.write(request.data)
      else:
        serialize_to(request, w)

      data = w.getvalue()
      f = Future(self.addr(), rpc_id)
      self._futures[rpc_id] = f
      #util.log_info('Send %s, %s', self.addr(), rpc_id)
      self._socket.send(data)
      return f

  def addr(self):
    return self._socket.addr

  def close(self):
    self._socket.close()

  def __getattr__(self, method_name):
    return ProxyMethod(self, method_name)

  def handle_read(self, socket):
    data = socket.recv()
    reader = cStringIO.StringIO(data)
    header = cPickle.load(reader)
    resp = read(reader)
    #resp = cPickle.load(reader)
    rpc_id = header['rpc_id']
    f = self._futures[rpc_id]
    f.done(resp)
    del self._futures[rpc_id]

  def close(self):
    self._socket.close()


def forall(clients, method, request):
  '''Invoke ``method`` with ``request`` for each client in ``clients``

  ``request`` is only serialized once, so this is more efficient when
  targeting multiple workers with the same data.

  Returns a future wrapping all of the requests.
  '''
  futures = []
  pickled = PickledData(data=serialize(request))
  for c in clients:
    futures.append(getattr(c, method)(pickled))

  return FutureGroup(futures)




import socket
import threading

import time
from spartan import util, rpc, core, blob_ctx


class Master(object):
  def __init__(self, port, num_workers):
    self._workers = {}
    self.num_workers = num_workers
    self._port = port
    self._server = rpc.listen(socket.gethostname(), port)
    self._server.register_object(self)
    self._initialized = False
    self._server.serve_nonblock()
    self._ctx = None

  def shutdown(self):
    self._ctx.active = False

    futures = rpc.FutureGroup()
    for id, w in self._workers.iteritems():
      util.log_info('Shutting down worker %d', id)
      futures.append(w.shutdown())
    self._server.shutdown()

  def register(self, req, handle):
    id = len(self._workers)
    self._workers[id] = rpc.connect(req.host, req.port)
    util.log_info('Registered worker %d of %d', id, self.num_workers)

    resp = core.RegisterResp()
    handle.done(resp)

    if len(self._workers) == self.num_workers:
      threading.Thread(target=self._initialize).start()

  def _initialize(self):
    util.log_info('Initializing...')


    req = core.Initialize(peers=dict([(id, w.addr())
                                      for id, w in self._workers.iteritems()]))

    futures = rpc.FutureGroup()
    for id, w in self._workers.iteritems():
      req.id = id
      futures.append(w.initialize(req))
    futures.wait()

    self._ctx = blob_ctx.BlobCtx(blob_ctx.MASTER_ID, self._workers)
    self._initialized = True

  def wait_for_initialization(self):
    while not self._initialized:
      time.sleep(0.1)

    blob_ctx.set(self._ctx)


'''Lazy arrays.

Expr operations are not performed immediately, but are set aside
and built into a control flow graph, which is then compiled
into a series of primitive array operations.
'''

import numpy as np

from ..node import Node
from .. import blob_ctx, util
from ..array import distarray


class NotShapeable(Exception):
  pass

unique_id = iter(xrange(10000000))

def _map(*args, **kw):
  '''
  Indirection for handling builtin operators (+,-,/,*).

  (Map is implemented in map.py)
  '''
  fn = kw['fn']
  numpy_expr = kw.get('numpy_expr', None)

  from .map import map
  return map(args, fn, numpy_expr)

class Expr(object):
  _cached_value = None
  _optimized = None
  _expr_id = None

  def dependencies(self):
    '''
    Return a dictionary mapping from name -> dependency.
    
    Dependencies may either be a list or single value.
    Dependencies of type `Expr` are recursively evaluated.
    '''
    return dict([(k, getattr(self, k)) for k in self.members])

  def compute_shape(self):
    '''
    Compute the shape of this expression.
    
    If the shape is not available (data dependent), raises `NotShapeable`.
    '''
    raise NotShapeable

  def visit(self, visitor):
    '''
    Apply visitor to all children of this node, returning a new `Expr` of the same type. 
    :param visitor: `OptimizePass`
    '''
    deps = {}
    for k in self.members:
      deps[k] = visitor.visit(getattr(self, k))

    return self.__class__(**deps)


  def node_init(self):
    #assert self.expr_id is not None
    if self._expr_id is None:
      self._expr_id = unique_id.next()

  def evaluate(self, ctx, deps):
    raise NotImplementedError

  def __hash__(self):
    return self._expr_id

  def typename(self):
    return self.__class__.__name__

  def __add__(self, other):
    return _map(self, other, fn=np.add, numpy_expr='+')

  def __sub__(self, other):
    return _map(self, other, fn=np.subtract, numpy_expr='-')

  def __mul__(self, other):
    return _map(self, other, fn=np.multiply, numpy_expr='*')

  def __mod__(self, other):
    return _map(self, other, fn=np.mod, numpy_expr='%')

  def __div__(self, other):
    return _map(self, other, fn=np.divide, numpy_expr='/')

  def __eq__(self, other):
    return _map(self, other, fn=np.equal, numpy_expr='==')

  def __ne__(self, other):
    return _map(self, other, fn=np.not_equal, numpy_expr='!=')

  def __lt__(self, other):
    return _map(self, other, fn=np.less, numpy_expr='<')

  def __gt__(self, other):
    return _map(self, other, fn=np.greater, numpy_expr='>')

  def __pow__(self, other):
    return _map(self, other, fn=np.power, numpy_expr='**')

  def __neg__(self):
    return _map(self, fn=np.negative)

  def __getitem__(self, idx):
    from .index import IndexExpr
    return IndexExpr(src=self, idx=lazify(idx))

  def __setitem__(self, k, val):
    raise Exception, 'Expressions are read-only.'

  @property
  def shape(self):
    '''Try to compute the shape of this DAG.
    
    If the value has been computed already this always succeeds.
    '''
    if self._cached_value is not None:
      return self._cached_value.shape

    try:
      return self.compute_shape()
    except NotShapeable:
      return evaluate(self).shape

  def force(self):
    return force(self)

  def dag(self):
    return dag(self)

  def glom(self):
    return glom(self)

  def __reduce__(self):
    return evaluate(self).__reduce__()

Expr.__rsub__ = Expr.__sub__
Expr.__radd__ = Expr.__add__
Expr.__rmul__ = Expr.__mul__
Expr.__rdiv__ = Expr.__div__


class AsArray(Expr):
  '''Promote a value to be array-like.

  This should be wrapped around most user-inputs that may be
  used in an array context, e.g. (``1 + x => map((as_array(1), as_array(x)), +)``)
  '''
  __metaclass__ = Node
  _members = ['val']

  def visit(self, visitor):
    return self

  def compute_shape(self):
    raise NotShapeable

  def evaluate(self, ctx, deps):
    util.log_info('Evaluate: %s', deps['val'])
    return distarray.as_array(deps['val'])

  def __str__(self):
    return 'V(%s)' % self.val


class LazyVal(Expr):
  __metaclass__ = Node
  _members = ['val']

  def visit(self, visitor):
    return self

  def dependencies(self):
    return {}

  def compute_shape(self):
    return self.val.shape

  def evaluate(self, ctx, deps):
    return self.val

  def __str__(self):
    return 'lazy(%s)' % self.val

class LazyCollection(Expr):
  '''
  LazyCollections wrap normal tuples, lists and dicts with `Expr` semantics.
  
  visit() and evaluate() are supported; these thread the visitor through
  child elements as expected.
  '''
  _members = ['vals']

  def __str__(self):
    return 'lazy(%s)' % (self.vals,)

  def evaluate(self, ctx, deps):
    return deps['vals']

  def __getitem__(self, idx):
    return self.vals[idx]

  def __iter__(self):
    return iter(self.vals)


class LazyDict(LazyCollection):
  __metaclass__ = Node
  def visit(self, visitor):
    return LazyDict(vals=dict([(k, visitor.visit(v)) for (k, v) in self.vals.iteritems()]))


class LazyList(LazyCollection):
  __metaclass__ = Node
  def visit(self, visitor):
    return LazyList(vals=[visitor.visit(v) for v in self.vals])


class LazyTuple(LazyCollection):
  __metaclass__ = Node
  def visit(self, visitor):
    return LazyTuple(vals=tuple([visitor.visit(v) for v in self.vals]))


def make_primitive(name, arg_names, evaluate_fn):
  class NewPrimitive(Expr):
    __metaclass__ = Node
    _members = arg_names

    def evaluate(self, ctx, deps):
      return evaluate_fn(ctx, **deps)


  NewPrimitive.__name__= name
  return NewPrimitive


def glom(node):
  '''
  Evaluate this expression and return the result as a `numpy.ndarray`. 
  '''
  if isinstance(node, Expr):
    node = evaluate(node)

  if isinstance(node, np.ndarray):
    return node

  return node.glom()


def dag(node):
  '''
  Compile and return the DAG representing this expression.
  
  :param node: The node to compute a DAG for.
  '''
  if not isinstance(node, Expr):
    raise TypeError

  if node._optimized is not None:
    return node._optimized

  from . import optimize
  dag = optimize.optimize(node)
  node._optimized = dag
  return node._optimized


def evaluate(node):
  '''
  Evaluate this expression.
  
  :param node: `Expr` to evaluate.
  '''
  if not isinstance(node, Expr):
    return node

  from . import backend
  result = backend.evaluate(blob_ctx.get(), dag(node))
  node._cached_value = result
  return result

force = evaluate

def eager(node):
  '''
  Eagerly evaluate ``node`` and convert the result back into an `Expr`.
  
  :param node: `Expr` to evaluate.
  '''
  return lazify(force(node))


def lazify(val):
  '''
  Lift ``val`` into an Expr node.
 
  If ``val`` is already an expression, it is returned unmodified.
   
  :param val:
  '''
  #util.log_info('Lazifying... %s', val)
  if isinstance(val, Expr):
    return val

  if isinstance(val, dict):
    return LazyDict(vals=val)

  if isinstance(val, list):
    return LazyList(vals=val)

  if isinstance(val, tuple):
    return LazyTuple(vals=val)

  return LazyVal(val=val)

def as_array(lst):
  results = []
  for v in lst:
    if isinstance(v, Expr):
      results.append(v)
    else:
      results.append(AsArray(v))

  return lazify(results)


import numpy as np
from scipy.sparse import dok_matrix
from . import extent
from spartan import util
from spartan.util import Assert


MASK_VALID = 1
MASK_INVALID = 0

TYPE_DENSE = 0
TYPE_SPARSE = 1


class Tile(object):
  '''
  A tile of an array: an extent (offset + size) and data for that extent.

  Tiles implement the Blob protocol: see `Blob`.
  '''

  def __init__(self, shape, dtype):
    self.shape = shape
    self.dtype = dtype

  @property
  def data(self):
    return self._data

  @data.setter
  def data(self, val):
    if val is not None:
      Assert.eq(val.dtype, self.dtype)

    self._data = val

  def get(self, selector):
    raise NotImplementedError

  def update(self, update, reducer):
    raise NotImplementedError


class DenseTile(Tile):
  '''Dense tiles use regular numpy arrays for storage.

  They track updates using a mask to determine valid regions.
  This masking is done manually, instead of using numpy.ma (MaskedArray),
  as MaskedArrays do not seem to properly handle masking of
  structured arrays.  Conversion to MaskedArrays is performed on
  fetch.
  '''

  def __init__(self, shape, dtype, data, mask=None):
    Tile.__init__(self, shape, dtype)

    if data is not None:
      Assert.eq(data.shape, shape)
      if mask is None:
        mask = MASK_VALID
    else:
      mask = MASK_INVALID

    self.mask = mask
    self.data = data
    self.shape = shape
    self.dtype = dtype

  def update(self, update, reducer):
    return merge_dense(self, update, reducer)

  def get(self, selector=None):
    self._initialize()

    # scalars are just returned directly.
    if len(self.data.shape) == 0:
      return self.data

    if selector is None or extent.is_complete(self.data.shape, selector):
      data = self.data
      mask = self.mask
    else:
      Assert.le(len(selector), len(self.data.shape),
                'Selector has more dimensions than data! %s %s' % (selector, self.data.shape))

      data = self.data[selector]
      mask = self.mask[selector]

    masked = np.ma.masked_all(data.shape, dtype=data.dtype)
    #util.log_info("\nM:%s\nD%s\n", data, mask)
    masked[mask] = data[mask]
    return masked

  def _initialize(self):
    # if this is a scalar, then data is just the scalar value itself.
    if len(self.shape) == 0:
      return

    if self.data is None:
      self.data = np.ndarray(self.shape, dtype=self.dtype)

    if not isinstance(self.mask, np.ndarray):
      mask_val = self.mask
      self.mask = np.ndarray(self.shape, dtype=np.bool)
      self.mask[:] = mask_val

  def __setitem__(self, idx, val):
    self._initialize()
    self.mask[idx] = MASK_VALID
    self.data[idx] = val

  def __repr__(self):
    return 'dense(%s, %s)' % (self.shape, self.dtype)


class SparseTile(Tile):
  def __init__(self, shape, data, dtype):
    Tile.__init__(self, shape, dtype)
    self.data = data

  def get(self, selector=None):
    self._initialize()
    assert selector is None or extent.is_complete(self.shape, selector)
    return self.data

  def _initialize(self):
    if self.data is None:
      self.data = dok_matrix(self.shape, self.dtype)

  def __setitem__(self, idx, val):
    self._initialize()
    self.data[idx] = val


def from_data(data):
  if isinstance(data, np.ndarray) or np.isscalar(data):
    return DenseTile(
      shape=data.shape,
      data=data,
      dtype=data.dtype)
  else:
    return SparseTile(shape=data.shape, data=data, dtype=data.dtype)


def from_shape(shape, dtype, tile_type=TYPE_DENSE):
  assert tile_type == TYPE_DENSE
  return DenseTile(shape=shape, data=None, dtype=dtype)


def from_intersection(src, overlap, data):
  '''
  Return a tile for ``src``, masked to update the area specifed by ``overlap``.
  
  :param src: `TileExtent`
  :param overlap: `TileExtent`
  :param data:
  '''
  slc = extent.offset_slice(src, overlap)
  tdata = np.ndarray(src.shape, data.dtype)
  tmask = np.ndarray(src.shape, np.bool)
  tmask[:] = MASK_INVALID
  tmask[slc] = MASK_VALID
  tdata[slc] = data
  return DenseTile(dtype=data.dtype, data=tdata, shape=src.shape, mask=tmask)


def merge_dense(old_tile, new_tile, reducer):
  Assert.isinstance(old_tile, Tile)
  Assert.isinstance(new_tile, Tile)

  Assert.eq(old_tile.dtype, new_tile.dtype)
  Assert.eq(old_tile.shape, new_tile.shape)

  if old_tile.data is None:
    return new_tile

  old_tile._initialize()
  new_tile._initialize()

  #util.log_info('OLD: %s', old_tile.data)
  #util.log_info('NEW: %s', new_tile.data)

  # zero-dimensional arrays; just use
  # data == None as a mask.
  if len(old_tile.shape) == 0:
    if old_tile.data is None or reducer is None:
      old_tile.data = new_tile.data
    else:
      old_tile.data = reducer(old_tile.data, new_tile.data)
    return old_tile

  Assert.eq(old_tile.shape, new_tile.shape)

  replaced = ~old_tile.mask & new_tile.mask
  updated = old_tile.mask & new_tile.mask

  old_tile.data[replaced] = new_tile.data[replaced]
  old_tile.mask[replaced] = MASK_VALID

  if np.any(updated):
    if reducer is None:
      old_tile.data[updated] = new_tile.data[updated]
    else:
      old_tile.data[updated] = reducer(old_tile.data[updated], new_tile.data[updated])

  return old_tile



from .base import Expr
import numpy as np
from spartan.node import Node


class OuterProductExpr(Expr):
  __metaclass__ = Node
  _members = ['children', 'map_fn', 'map_fn_kw', 'reduce_fn', 'reduce_fn_kw']
  
def outer_product(a, b, map_fn, reduce_fn):
  '''
  Outer (cartesian) product over the tiles of ``a`` and ``b``.
  
  ``map_fn`` is applied to each pair; ``reduce_fn`` is used to 
  combine overlapping outputs.
  
  :param a:
  :param b:
  '''
  return OuterProductExpr(a, b, map_fn, reduce_fn)

def outer(a, b):
  
  return OuterProductExpr(a, b, map_fn=np.dot, 
                          reduce_fn=np.add)



#!/usr/bin/env python

from contextlib import contextmanager
import logging
from math import ceil
from os.path import basename
import collections
import os
import select
import sys
import threading
import time
import traceback
import cStringIO

import numpy as np


log_debug = logging.debug
log_info = logging.info
log_warn = logging.warn
log_error = logging.error
log_fatal = logging.fatal

def findCaller(obj):
  f = sys._getframe(4)
  co = f.f_code
  filename = os.path.normcase(co.co_filename)
  return co.co_filename, f.f_lineno, co.co_name

root = logging.getLogger()
logging.RootLogger.findCaller = findCaller


class FileWatchdog(threading.Thread):
  """Watchdog for a file (typically `sys.stdin` or `sys.stdout`).

  When the file closes, terminate the process.
  (This occurs when an ssh connection is terminated, for example.)
  """
  def __init__(self, file_handle=sys.stdin, on_closed=lambda: os._exit(1)):
    '''
    
    :param file_handle:
    :param on_closed:
    '''
    threading.Thread.__init__(self, name='WatchdogThread')
    self.setDaemon(True)
    self.file_handle = file_handle
    self.on_closed = on_closed

  def run(self):
    f = [self.file_handle]
    while 1:
      r, w, x = select.select(f, f, f, 1.0)
      if r:
        self.on_closed()
        return
      
      time.sleep(0.1)

def flatten(lst, depth=1):
  if depth == 0:
    return lst

  out = []
  for item in lst:
    if isinstance(item, (list, set)): out.extend(flatten(item, depth=depth - 1))
    else: out.append(item)
  return out


def timeit(f, name=None):
  '''
  Run ``f`` and log the amount of time taken.
   
  :param f:
  :param name:
  '''
  st = time.time()
  res = f()
  ed = time.time()
  if name is None:
    name = f
  log_info('Operation %s completed in %.3f seconds', name, ed - st)
  return res

@contextmanager
def timer_ctx(name='Operation'):
  '''Context based timer:

  Usage::

    with timer_ctx('LoopOp'):
      for i in range(10):
        my_op()

  '''
  st = time.time()
  yield
  ed = time.time()
  log_info('%3.5f:: %s' % (ed - st, name))

class EZTimer(object):
  '''Lazy timer.

  Prints elapsed time when destroyed.
  '''
  def __init__(self, name):
    self.name = name
    self.st = time.time()

  def __del__(self):
    print('%3.5f:: %s' % (time.time() - self.st, self.name))

class Timer(object):
  def __init__(self):
    self.elapsed = 0

  def start(self):
    self.st = time.time()

  def stop(self):
    self.elapsed += time.time() - self.st

  def __enter__(self):
    self.start()

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.stop()

def dump_stacks(out):
  '''Dump the stacks of all threads.'''
  id_to_name = dict([(th.ident, th.name) for th in threading.enumerate()])
  thread_stacks = collections.defaultdict(list)

  for thread_id, stack in sys._current_frames().items():
    code = []
    for filename, lineno, name, line in traceback.extract_stack(stack):
      if line is None: line = ''
      code.append('%s:%d (%s): %s' % (basename(filename), lineno, name, line.strip()))
    thread_stacks['\n'.join(code)].append(thread_id)

  for stack, thread_ids in thread_stacks.iteritems():
    print >> out, 'Thread %d(%s)' % (thread_ids[0], id_to_name.get(thread_ids[0], ''))
    if len(thread_ids) > 1:
      print >> out, '... and %d more' % (len(thread_ids) - 1)
    print >> out, stack
    print >> out

def stack_signal():
  out = cStringIO.StringIO()
  dump_stacks(out)
  print >> sys.stderr, out.getvalue()
  with open('/tmp/%d.stacks' % os.getpid(), 'w') as f:
    print >> f, out.getvalue()


class Assert(object):
  '''Assertion helper functions.
  
  ::
  
    a = 'foo'
    b = 'bar'
    
    Assert.eq(a, b) 
    # equivalent to:
    # assert a == b, 'a == b failed (%s vs %s)' % (a, b) 
  '''
  @staticmethod
  def all_eq(a, b):
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
      assert a.shape == b.shape, 'Mismatched shapes: %s %s' % (a.shape, b.shape)
      assert np.all(a == b), 'Failed: \n%s\n ==\n%s' % (a, b)
      return

    if np.isscalar(a) or np.isscalar(b):
      assert a == b, 'Failed: \n%s\n ==\n%s' % (a, b)
      return

    assert iterable(a), (a, b)
    assert iterable(b), (a, b)

    for i, j in zip(a, b):
      assert i == j, 'Failed: \n%s\n ==\n%s' % (a, b)

  @staticmethod
  def eq(a, b, msg=''): 
    assert (a == b), 'Failed: %s == %s (%s)' % (a, b, msg)
  
  @staticmethod
  def ne(a, b): assert (a == b), 'Failed: %s != %s' % (a, b)
  
  @staticmethod
  def gt(a, b): assert (a > b), 'Failed: %s > %s' % (a, b)
  
  @staticmethod
  def lt(a, b): assert (a < b), 'Failed: %s < %s' % (a, b)
  
  @staticmethod
  def ge(a, b): assert (a >= b), 'Failed: %s >= %s' % (a, b)
  
  @staticmethod
  def le(a, b, msg=None):
    if msg is None:
      assert (a <= b), 'Failed: %s <= %s' % (a, b)
    else:
      assert (a <= b), 'Failed: %s <= %s [%s]' % (a, b, msg)

  @staticmethod
  def true(expr): assert expr, 'Failed: %s == True' % (expr)
 
  @staticmethod
  def iterable(expr): assert iterable(expr), 'Not iterable: %s' % expr
   
  @staticmethod
  def isinstance(expr, klass): 
    assert isinstance(expr, klass), 'Failed: isinstance(%s, %s) [type = %s]' % (expr, klass, type(expr))
  
  @staticmethod
  def no_duplicates(collection):
    d = collections.defaultdict(int)
    for item in collection:
      d[item] += 1
    
    bad = [(k,v) for k, v in d.iteritems() if v > 1]
    assert len(bad) == 0, 'Duplicates found: %s' % bad
  
 
def trace_fn(fn):
  '''Function decorator: log on entry and exit to ``fn``.'''
  def tracer(*args, **kw):
    log_info('TRACE: >> %s with args: %s %s', fn, args, kw)
    result = fn(*args, **kw)
    log_info('TRACE: << %s (%s)', fn, result)
    return result
  return tracer
   
   
def rtype_check(typeclass):
  '''Function decorator to check return type.
  
  Usage::
  
    @rtype_check(int)
    def fn(x, y, z):
      return x + y
  '''
  
  def wrap(fn):
    def checked_fn(*args, **kw):  
      result = fn(*args, **kw)
      Assert.isinstance(result, typeclass)
      return result
    checked_fn.__name__ = 'checked_' + fn.__name__
    checked_fn.__doc__ = fn.__doc__
    return checked_fn
  return wrap

def synchronized(fn):
  '''
  Decorator: execution of this function is serialized by an `threading.RLock`.
  :param fn:
  '''
  lock = threading.RLock()
  def _fn(*args, **kw):
    with lock:
      return fn(*args, **kw)

  if hasattr(fn, '__name__'):
    _fn.__name__ = fn.__name__
  else:
    _fn.__name__ = 'unnamed'

  return _fn
  
def count_calls(fn):
  '''
  Decorator: count calls to ``fn`` and print after each 100.
  :param fn:
  '''
  count = [0]
  def wrapped(*args, **kw):
    count[0] += 1
    if count[0] % 100 == 0: print count[0], fn.__name__
    return fn(*args, **kw)
  
  wrapped.__name__ = 'counted_' + fn.__name__
  wrapped.__doc__ = fn.__doc__
  return wrapped

def join_tuple(tuple_a, tuple_b):
  return tuple(list(tuple_a) + list(tuple_b))

def divup(a, b):
  if isinstance(a, tuple):
    return tuple([divup(ta, b) for ta in a])
  
  return int(ceil(float(a) / b))

def iterable(x):
  return hasattr(x, '__iter__')

def as_list(x):
  if isinstance(x, list): return x
  if iterable(x): return list(x)
  
  return [x]


def get_core_mapping():
  lines = open('/proc/cpuinfo').read().strip().split('\n')
  package_id = core_id = None
  id = 0
  cpus = {}
  for l in lines:
    if l.startswith('physical id'):
      package_id = int(l.split(':')[1].strip())

    if l.startswith('core_id'):
      core_id = int(l.split(':')[1].strip())

    if package_id is not None and core_id is not None:
      cpus[id] = (package_id, core_id)

  return cpus

def get_good_cores():
  """
  Return a list of processor ids that correspond to the primary thread on each core.
  :return:
  """
  cpus = get_core_mapping()
  unique = {}
  m



'''
Indexing operations (slicing and filtering).
'''
from .base import Expr, LazyList
from spartan import util
from spartan.array import extent, tile, distarray
from spartan.node import Node
from spartan.util import Assert, join_tuple
import numpy as np


class IndexExpr(Expr):
  __metaclass__ = Node
  _members = ['src', 'idx']

  def node_init(self):
    Expr.node_init(self)
    assert not isinstance(self.src, LazyList)
    assert not isinstance(self.idx, LazyList)
    assert not isinstance(self.idx, list)

  def evaluate(self, ctx, deps):
    idx = deps['idx']
    assert not isinstance(idx, list) 
    util.log_info('Indexing: %s', idx)
    if isinstance(idx, tuple) or\
       isinstance(idx, slice) or\
       np.isscalar(idx):
      return eval_Slice(ctx, self, deps)
    
    return eval_Index(ctx, self, deps)
     

def int_index_mapper(ex, src, idx, dst):
  '''Map over the index array, fetching rows from the data array.'''
  idx_vals = idx.fetch(extent.drop_axis(ex, -1))
  tile = src.fetch(ex)
  
  util.log_info('Dest shape: %s, idx: %s, %s', tile.shape, ex, idx_vals)
  for dst_idx, src_idx in enumerate(idx_vals):
    tile[dst_idx] = src.fetch(extent.from_slice(int(src_idx), src.shape))

  return [(ex, tile)]

def bool_index_mapper(ex, src, idx):
  val = src.fetch(ex)
  mask = idx.fetch(ex)

  #util.log_info('\nVal: %s\n Mask: %s', val, mask)
  return [(ex, np.ma.masked_array(val, mask))]

def eval_Index(ctx, prim, deps):
  src = deps['src']
  idx = deps['idx']
  
  Assert.isinstance(idx, (np.ndarray, distarray.DistArray))

  if idx.dtype == np.bool:
    # return a new array masked by `idx`
    dst = src.map_to_array(bool_index_mapper, kw={ 'src' : src, 'idx' : idx})
    return dst
  else:
    util.log_info('Integer indexing...')

    Assert.eq(len(idx.shape), 1)

    # create empty destination of same first dimension as the index array
    dst = distarray.create(join_tuple([idx.shape[0]], src.shape[1:]), dtype=src.dtype)
    
    # map over it, fetching the appropriate values for each tile.
    return dst.map_to_array(int_index_mapper, kw={ 'src' : src, 'idx' : idx, 'dst' : dst })

    
def eval_Slice(ctx, prim, deps):
  src = deps['src']
  idx = deps['idx']
  
  return distarray.Slice(src, idx)
  # return _eager_slice(src, idx)





#!/usr/bin/env python

'''
Implementation of the generic loop() expression.

:: 
  
  loop([xrange(start1, stop2),
        xrange(start2, stop2),
         ... ], fn)

The loop function should expect as input the 
current indices being operated on: e.g. ``fn(i, j)``.
'''

import itertools

import numpy as np
from spartan import util
from spartan.array import distarray
from spartan.expr.base import Expr
from spartan.node import Node


def _loop_kernel(kernel):
  mapper = kernel.arg('_args')
  args = kernel.arg('_workitem')
  return mapper(*args)

class LoopExpr(Expr):
  __metaclass__ = Node
  _members = ['ranges', 'sources', 'arg_fn', 'mapper_fn', 'target']
  
  def evaluate(self, ctx, deps):
    fn = deps['arg_fn']
    sources = deps['sources']
    worklist = []
    for i in itertools.product(*deps['ranges']):
      args = fn(*i)
      ex = args[0]
      worklist.append((args, distarray.best_locality(sources[0], ex)))

    kernel_args = deps['mapper_fn']
    ctx.foreach_worklist(_loop_kernel, kernel_args, worklist)
    return deps['target']
    
def loop(ranges, sources, arg_fn, mapper_fn, target):
  return LoopExpr(ranges=ranges,
                  sources=sources,
                  arg_fn=arg_fn,
                  mapper_fn=mapper_fn,
                  target=target)


'''ZeroMQ socket implementation.'''

import cProfile
import collections
import os
import threading
import socket
import traceback
import sys
import fcntl
import zmq

from .common import Group, SocketBase
from spartan import util

POLLER = None
POLLER_LOCK = threading.Lock()
PROFILER = None

def poller():
  global POLLER
  with POLLER_LOCK:
    if POLLER is None:
      #util.log_info('Started poller.. %s %s', os.getpid(), __file__)
      POLLER = ZMQPoller()
      POLLER.start()
    return POLLER

socket_count = collections.defaultdict(int)

class Socket(SocketBase):
  __slots__ = ['_zmq', '_hostport', '_out', '_in', '_addr', '_closed', '_shutdown', '_lock']

  def __init__(self, ctx, sock_type, hostport):
    socket_count[os.getpid()] += 1
    #util.log_info('New socket... %s.%s -> %s [%s]',
    #              socket.gethostname(), os.getpid(),  hostport, socket_count[os.getpid()])
    #s = ''.join(traceback.format_stack())
    #print >>sys.stderr, s
    self._zmq = ctx.socket(sock_type)
    self.addr = hostport
    self._out = collections.deque()
    self._closed = True
    self._shutdown = False
    self._lock = threading.RLock()

  def in_poll_loop(self):
    return threading.current_thread() == poller()

  def __repr__(self):
    return 'Socket(%s)' % ((self.addr,))

  def close(self, *args):
    if self._closed:
      return

    with self._lock:
      self._closed = True
      self._shutdown = True
      poller().close(self)

  def send(self, msg):
    assert not self._closed
    with self._lock:
      #util.log_info('SEND %s', len(msg))
      self._out.append(msg)
      poller().modify(self, zmq.POLLIN | zmq.POLLOUT)

  def zmq(self):
    return self._zmq

  def recv(self):
    assert not self._closed
    with self._lock:
      frames = self._zmq.recv_multipart(copy=False, track=False)
      assert len(frames) == 1
      return frames[0]

  def connect(self):
    assert self._closed
    with self._lock:
      #util.log_info('Connecting: %s:%d' % self.addr)
      self._zmq.connect('tcp://%s:%s' % self.addr)
      poller().add(self, zmq.POLLIN)
      self._closed = False

  @property
  def port(self):
    return self.addr[1]

  @property
  def host(self):
    return self.addr[0]

  def handle_close(self):
    self._zmq.close()
    self._zmq = None

  def handle_write(self):
    with self._lock:
      if not self._out:
        return

      while self._out:
        next = self._out.popleft()
        if isinstance(next, Group):
          #util.log_info('Sending group. %s', len(next))
          self._zmq.send_multipart(next, copy=False)
        else:
          #util.log_info('Sending %s', len(next))
          self._zmq.send(next, copy=False)

      poller().modify(self, zmq.POLLIN)

  def handle_read(self, socket):
    self._handler(socket)


class ServerSocket(Socket):
  def __init__(self, ctx, sock_type, hostport):
    Socket.__init__(self, ctx, sock_type, hostport)
    self.addr = hostport

  def send(self, msg):
    '''Send ``msg`` to a remote client.

    :param msg: `.Group`, with the first element being the destination to send to.
    '''
    Socket.send(self, msg)

  def bind(self):
    assert self._closed
    self._closed = False
    host, port = self.addr
    host = socket.gethostbyname(host)
    util.log_info('Binding... %s', (host, port))
    if port == -1:
      self.addr = (host, self._zmq.bind_to_random_port('tcp://%s' % host))
    else:
      try:
        self._zmq.bind('tcp://%s:%d' % (host, port))
      except zmq.ZMQError:
        util.log_info('Failed to bind (%s, %d)' % (host, port))
        raise

    poller().add(self, zmq.POLLIN)

  def handle_read(self, socket):
    packet = self._zmq.recv_multipart(copy=False, track=False)
    source, rest = packet[0], packet[1:]
    stub_socket = StubSocket(source, self, rest)
    self._handler(stub_socket)

  def zmq(self):
    return self._zmq


class StubSocket(SocketBase):
  '''Handles a single read from a client'''

  def __init__(self, source, socket, data):
    self._out = collections.deque()
    self.source = source
    self.socket = socket

    assert isinstance(data, list)
    assert len(data) == 1
    self.data = data[0]

  @property
  def addr(self):
    return self.socket.addr

  def recv(self):
    assert self.data is not None, 'Tried to read twice from stub socket.'
    res = self.data
    self.data = None
    # print 'Result!', res
    return res

  def send(self, req):
    if isinstance(req, Group):
      req = Group([self.source] + list(req))
    else:
      req = Group((self.source, req))
    self.socket.send(req)


class ZMQPoller(threading.Thread):
  def __init__(self):
    threading.Thread.__init__(self, name='zmq.PollingThread', target=self._run)

    self._poller = zmq.Poller()
    self._lock = threading.RLock()

    self._pipe = os.pipe()
    fcntl.fcntl(self._pipe[0], fcntl.F_SETFL, os.O_NONBLOCK)
    fcntl.fcntl(self._pipe[1], fcntl.F_SETFL, os.O_NONBLOCK)

    self._poller.register(self._pipe[0], zmq.POLLIN)
    self._sockets = {}
    self.profiler = cProfile.Profile()

    self._running = False

    self._to_add = []
    self._to_close = []
    self._to_mod = []
    self.daemon = True

  def _run(self):
    self._running = True
    _poll = self._poller.poll
    _poll_time = 1
    MAX_TIMEOUT = 100

    while self._running:
      self.profiler.disable()
      socks = dict(_poll(_poll_time))

      if len(socks) == 0:
        _poll_time = min(_poll_time * 2, MAX_TIMEOUT)
      else:
        _poll_time = 1

      self.profiler.enable()
      #util.log_info('%s', self._sockets)
      for fd, event in socks.iteritems():
        if fd == self._pipe[0]:
          os.read(fd, 10000)
          continue

        if not fd in self._sockets:
          continue

        socket = self._sockets[fd]
        if event & zmq.POLLIN: socket.handle_read(socket)
        if event & zmq.POLLOUT: socket.handle_write()

      with self._lock:
        for s, dir in self._to_add:
          self._sockets[s.zmq()] = s
          self._poller.register(s.zmq(), dir)

        for s, dir in self._to_mod:
          self._poller.register(s.zmq(), dir)

        for s in self._to_close:
          self._poller.unregister(s.zmq())
          del self._sockets[s.zmq()]
          socket.handle_close()

        del self._to_mod[:]
        del self._to_add[:]
        del self._to_close[:]

    util.log_info("ZMQ poller stopped.")

  def stop(self):
    if not self._running:
      return

    self._running = False
    self.wakeup()
    if threading.current_thread() != self:
      self.join()

  def wakeup(self):
    os.write(self._pipe[1], 'x')

  def modify(self, socket, direction):
    with self._lock:
      assert socket.zmq() in self._sockets
      self._to_mod.append((socket, direction))
      self.wakeup()

  def add(self, socket, direction):
    with self._lock:
      self._to_add.append((socket, direction))
      self.wakeup()

  def close(self, socket):
    'Execute socket.handle_close() from within the polling thread.'
    with self._lock:
      assert socket.zmq() in self._sockets
      self._to_close.append(socket)
      self.wakeup()

def shutdown():
  poller().stop()

def server_socket(addr):
  host, port = addr
  return ServerSocket(zmq.Context.instance(), zmq.ROUTER, (host, port))


def client_socket(addr):
  host, port = addr
  return Socket(zmq.Context.instance(), zmq.DEALER, (host, port))



'''Helper for constructing trees of objects.

Provides pretty printing, equality testing, hashing and keyword initialization.
'''  


import numpy as np
from itertools import izip
import copy

class NodeTemplate(object):
  def __init__(self, *args, **kw):
    n_args = len(args)
    members = self.members
    n_members = len(members)
    class_name = self.__class__.__name__
    self_dict = self.__dict__
    if n_args == n_members:
      assert len(kw) == 0
      for (k,v) in zip(members,args):
        self_dict[k] = v
    elif n_args < n_members:
      for field in members:
        self_dict[field] = kw.get(field)

      for field, value in izip(members, args):
        self_dict[field] = value

      for (k,v) in kw.iteritems():
        assert k in members, \
          "Keyword argument '%s' not recognized for %s: %s" % \
          (k, self.node_type(), members)
    else:
      raise Exception('Too many arguments for %s, expected %s' % \
                      (class_name, members))

    # it's more common to not define a node initializer,
    # so add an extra check to avoid having to always
    # traverse the full class hierarchy
    if hasattr(self, 'node_init'):
      for C in self.__class__.mro():
        if 'node_init' in C.__dict__:
          C.node_init(self)

  def __hash__(self):
    # print "Warning: __hash__ not implemented for %s" % self
    hash_value = 0x123123
    for m in self.members:
      v = getattr(self, m)
      if isinstance(v, np.ndarray):
        continue
      elif isinstance(v, (list, tuple)):
        for i in v: hash_value ^= hash(i)
      else:
        hash_value ^= hash(v)
    return hash_value

  def eq_members(self, other):
    print 'EQ:', self.__class__.__name__
    for (k,v) in self.iteritems():
      if not hasattr(other, k):
        return False
      if getattr(other, k) != v:
        return False
    return True

  def __eq__(self, other):
    return other.__class__ is  self.__class__ and self.eq_members(other)

  def __ne__(self, other):
    return not self == other

  def node_type(self):
    return self.__class__.__name__

  def clone(self, **kwds):
    cloned = copy.deepcopy(self)
    for (k,v) in kwds.values():
      setattr(cloned, k, v)
    return cloned

  def __str__(self):
    member_strings = []
    for (k,v) in self.iteritems():
      if isinstance(v, list):
        v_str = ['[']
        for i, v in enumerate(v):
          v_str.append('[%d] = %s' % (i, v))
        v_str += [']']

        v_str = '\n'.join(v_str)
      else:
        v_str = str(v)
      member_strings.append("%s = %s" % (k, v_str))
    child_str = '  ' + ',\n'.join(member_strings)
    child_str = child_str.replace('\n', '\n  ')

    return "%s { \n%s \n}" % (self.node_type(), child_str)

  def __repr__(self):
    return self.__str__()

  def iteritems(self):
    for k in self.members:
      yield (k, getattr(self, k, None))

  def itervalues(self):
    for (_,v) in self.iteritems():
      yield v

  def items(self):
    return [(k,getattr(self,k)) for k in self.members]


class Node(object):
  @classmethod
  def members(cls):
    return cls.members

  def __new__(cls, name, bases, attrs):
    members = set()
    for m in attrs.get('_members', []):
      members.add(m)

    for b in bases:
      for m in getattr(b, '_members', []):
        members.add(m)
    members = list(members)

    attrs['members'] = members
    attrs['__doc__'] = '\n'.join([':attribute %s: (initialized by Node)' % n for n in members])

    bases = list(bases)
    if object in bases:
      bases = [b for b in bases if b is not object]

    bases += [NodeTemplate]
    bases = tuple(bases)
    return type(name, bases, attrs)





#!/usr/bin/env python

import collections
import numpy as np

from spartan import util, blob_ctx
from spartan.array import distarray, tile
from spartan.node import Node
from spartan.util import Assert
from .base import Expr, lazify, as_array


def map(inputs, fn, numpy_expr=None, **kw):
  '''
  Evaluate ``fn`` over each tile of the input.
  
  ``fn`` should be of the form ``([inputs], **kw)``.
  :param v: `Expr`
  :param fn: callable taking arguments ``(inputs, **kw)``
  '''
  if not util.iterable(inputs):
    inputs = [inputs]

  inputs = as_array(inputs)
  kw = lazify(kw)
  
  return MapExpr(children=inputs, map_fn=fn, numpy_expr=None, fn_kw=kw)


def tile_mapper(ex, children, map_fn, fn_kw):
  ctx = blob_ctx.get()
  #util.log_info('MapTiles: %s', map_fn)
  #util.log_info('Fetching %d inputs', len(children))
  #util.log_info('%s %s', children, ex)
  local_values = [c.fetch(ex) for c in children]
  result = map_fn(*local_values, **fn_kw)
  #util.log_info('Result: %s', result)
  Assert.eq(ex.shape, result.shape, 'Bad shape: (%s)' % map_fn)
  return [(ex, result)]
    

class MapExpr(Expr):
  __metaclass__ = Node
  _members = ['children', 'map_fn', 'fn_kw', 'numpy_expr', 'local_dag']

  def __str__(self):
    children = ','.join([str(c) for c in self.children])
    children = children.replace('\n', '\n  ')
    return 'map(%s, \n  %s)' % (self.map_fn.__name__, children)
  
  def compute_shape(self):
    '''MapTiles retains the shape of inputs.
    
    Broadcasting results in a map taking the shape of the largest input.
    '''
    shapes = [i.shape for i in self.children]
    output_shape = collections.defaultdict(int)
    for s in shapes:
      for i, v in enumerate(s):
        output_shape[i] = max(output_shape[i], v)
    return tuple([output_shape[i] for i in range(len(output_shape))])
  
  def evaluate(self, ctx, deps):
    children = deps['children']
    map_fn = self.map_fn
    fn_kw = deps['fn_kw']

    if 'numpy_expr' in fn_kw:
      del fn_kw['numpy_expr']

    assert fn_kw is not None

    #for c in children:
    #  util.log_info('Child: %s', c)

    children = distarray.broadcast(children)
    largest = distarray.largest_value(children)
    util.log_info('Mapping %s over %d inputs; largest = %s', map_fn, len(children), largest.shape)

    result = largest.map_to_array(tile_mapper,
                                  kw = { 'children' : children,
                                         'map_fn' : map_fn,
                                         'fn_kw' : fn_kw })
    return result


#!/usr/bin/env python

import itertools
import collections

import numpy as np

from . import tile, extent
from spartan import util, core, blob_ctx, rpc
from spartan.util import Assert


# number of elements per tile
TILE_SIZE = 100000


def take_first(a,b):
  return a


def compute_splits(shape, tile_hint=None, num_shards=-1):
  '''Split an array of shape ``shape`` into `Extent`s containing roughly `TILE_SIZE` elements.
 
  :param shape: tuple
  :param tile_hint: tuple indicating the desired tile shape 
  :rtype: list of `Extent`
  '''
  
  splits = [None] * len(shape)
  if tile_hint is None:
    # try to make reasonable tiles
    if len(shape) == 0:
      return { extent.create([], [], ()) :  0 }
   
    weight = 1
    
    # split each dimension into tiles.  the first dimension
    # is kept contiguous if possible.
    for dim in reversed(range(len(shape))):
      step = max(1, TILE_SIZE / weight)
      dim_splits = []
      for i in range(0, shape[dim], step):
        dim_splits.append((i, min(shape[dim], i + step)))
        
      splits[dim] = dim_splits
      weight *= shape[dim]
  else:
    Assert.eq(len(tile_hint), len(shape),
              '#dimensions in tile hint does not match shape %s vs %s' % (tile_hint, shape))
    for dim in range(len(shape)):
      dim_splits = []
      step = tile_hint[dim]
      #Assert.le(step, shape[dim])
      for i in range(0, shape[dim], step):
        dim_splits.append((i, min(shape[dim],  i + step)))
      splits[dim] = dim_splits
 
  result = {}
  idx = 0
  for slc in itertools.product(*splits):
    if num_shards != -1:
      idx = idx % num_shards
      
    ul, lr = zip(*slc)
    ex = extent.create(ul, lr, shape)
    result[ex] = idx
    idx += 1
  
  return result
    

def _array_mapper(blob_id, blob, array=None, user_fn=None, **kw):
  ctx = blob_ctx.get()
  ex = array.extent_for_blob(blob_id)
  results = []
  user_results = user_fn(ex, **kw)
  assert user_results is not None, user_fn

  for target_ex, target_data in user_results:
    Assert.eq(target_ex.shape, target_data.shape,
              'Bad extent for result: %s %s' % (user_fn, ex))
    blob_id = ctx.create(tile.from_data(target_data)).wait().blob_id
    results.append((target_ex, blob_id))
  return results

class DistArray(object):
  '''The interface required for distributed arrays.'''

  def fetch(self, ex):
    raise NotImplementedError

  def update(self, ex, data):
    raise NotImplementedError

  def __repr__(self):
    return '%s(shape=%s, dtype=%s)' % (self.__class__.__name__, self.shape, self.dtype)

  def select(self, idx):
    '''
    Effectively __getitem__.

    Renamed to avoid the chance of accidentally using a slow, local operation on
    a distributed array.
    '''
    if isinstance(idx, extent.TileExtent):
      return self.fetch(idx)

    if np.isscalar(idx):
      return self[idx:idx+1][0]

    ex = extent.from_slice(idx, self.shape)
    #util.log_info('Select: %s + %s -> %s', idx, self.shape, ex)
    return self.fetch(ex)

  def glom(self):
    #util.log_info('Glomming: %s', self.shape)
    return self.select(np.index_exp[:])

  def map_to_array(self, mapper_fn, kw=None):
    results = self.map_to_table(mapper_fn=mapper_fn, kw=kw)
    extents = {}
    for blob_id, d in results.iteritems():
      for ex, id in d:
        extents[ex] = id

    return from_table(extents)

  def foreach(self, mapper_fn, kw):
    return self.map_to_table(mapper_fn=mapper_fn, kw=kw)

ID_COUNTER = iter(xrange(10000000))

class DistArrayImpl(DistArray):
  def __init__(self, shape, dtype, tiles, reducer_fn):
    self.shape = shape
    self.dtype = dtype
    self.reducer_fn = reducer_fn
    self.ctx = blob_ctx.get()

    #util.log_info('%s', extents)
    Assert.isinstance(tiles, dict)

    self.blob_to_ex = {}
    for k,v in tiles.iteritems():
      Assert.isinstance(k, extent.TileExtent)
      Assert.isinstance(v, core.BlobId)
      self.blob_to_ex[v] = k

    self.tiles = tiles
    self.id = ID_COUNTER.next()

  def __reduce__(self):
    return (DistArrayImpl, (self.shape, self.dtype, self.tiles, self.reducer_fn))

  def __del__(self):
    '''Destroy this array.

    NB: Destruction is actually deferred until the next usage of the
    blob_ctx.  __del__ can be called at anytime, including the
    invocation of a RPC call, which leads to odd/bad behavior.
    '''
    if self.ctx.worker_id == blob_ctx.MASTER_ID:
      #util.log_info('Destroying table... %s', self.id)
      ctx = self.ctx
      tiles = self.tiles.values()

      # drop reference to self from lambda
      self.ctx.defer(lambda: ctx.destroy_all(tiles))

  def id(self):
    return self.table.id()

  def extent_for_blob(self, id):
    return self.blob_to_ex[id]
  
  def tile_shape(self):
    scounts = collections.defaultdict(int)
    for ex in self.tiles.iterkeys():
      scounts[ex.shape] += 1
    
    return sorted(scounts.items(), key=lambda kv: kv[1])[-1][0]

  def map_to_table(self, mapper_fn, kw=None):
    ctx = blob_ctx.get()

    if kw is None: kw = {}
    kw['array'] = self
    kw['user_fn'] = mapper_fn

    return ctx.map(self.tiles.values(),
                   mapper_fn = _array_mapper,
                   reduce_fn=None,
                   kw=kw)

  def fetch(self, region):
    '''
    Return a local numpy array for the given region.
    
    If necessary, data will be copied from remote hosts to fill the region.    
    :param region: `Extent` indicating the region to fetch.
    '''
    Assert.isinstance(region, extent.TileExtent)
    Assert.eq(region.array_shape, self.shape)
    Assert.eq(len(region.ul), len(self.shape))

    #util.log_info('FETCH: %s %s', self.shape, region)

    ctx = blob_ctx.get()
    assert np.all(region.lr <= self.shape), (region, self.shape)
    
    # special case exact match against a tile 
    if region in self.tiles:
      #util.log_info('Exact match.')
      ex, intersection = region, region
      blob_id = self.tiles[region]
      tgt = ctx.get(blob_id, extent.offset_slice(ex, intersection))
    else:
      splits = list(extent.find_overlapping(self.tiles.iterkeys(), region))

      #util.log_info('Target shape: %s, %d splits', region.shape, len(splits))
      tgt = np.ma.MaskedArray(np.ndarray(region.shape, dtype=self.dtype))
      tgt.mask = tile.MASK_INVALID
      #util.log_info('Fetching %d tiles', len(splits))

      futures = []
      for ex, intersection in splits:
        dst_slice = extent.offset_slice(region, intersection)
        blob_id = self.tiles[ex]

        def _apply_data(r, slc=dst_slice):
          tgt[slc] = r.data

        futures.append(ctx.get(blob_id, extent.offset_slice(ex, intersection),
                               callback=_apply_data))
        #tgt[dst_slice] = ctx.get(blob_id, extent.offset_slice(ex, intersection))

        #util.log_info('%s %s', dst_slice, src_slice.shape)
      rpc.wait_for_all(futures)

    # attempt to remove mask on arrays when it is
    # all valid.
    if isinstance(tgt, np.ma.MaskedArray) and np.all(tgt.mask == np.ma.nomask):
      return tgt.data

    return tgt
    #return tile.data[]
   
  def update_slice(self, slc, data):
    return self.update(extent.from_slice(slc, self.shape), data)
     
  def update(self, region, data):
    ctx = blob_ctx.get()
    Assert.isinstance(region, extent.TileExtent)
    Assert.eq(region.shape, data.shape,
              'Size of extent does not match size of data')

    #util.log_info('%s %s', self.table.id(), self.tiles)
    # exact match
    if region in self.tiles:
      #util.log_info('EXACT: %d %s ', self.table.id(), region)
      blob_id = self.tiles[region]
      ctx.update(blob_id, tile.from_data(data), self.reducer_fn)
      return
    
    splits = list(extent.find_overlapping(self.tiles, region))
    futures = []
    #util.log_info('Updating %s tiles', len(splits))
    for dst_key, intersection in splits:
      #util.log_info('%d %s %s %s', self.table.id(), region, dst_key, intersection)
      blob_id = self.tiles[dst_key]
      src_slice = extent.offset_slice(region, intersection)
      update_tile = tile.from_intersection(dst_key, intersection, data[src_slice])
      #util.log_info('%s %s %s %s', dst_key.shape, intersection.shape, blob_id, update_tile)
      futures.append(ctx.update(blob_id, update_tile, self.reducer_fn, wait=False))

    rpc.wait_for_all(futures)


def create(shape,
           dtype=np.float,
           sharder=None,
           combiner=None,
           reducer=None,
           tile_hint=None):
  '''Make a new, empty DistArray'''
  ctx = blob_ctx.get()
  dtype = np.dtype(dtype)
  shape = tuple(shape)

  extents = compute_splits(shape, tile_hint, -1).keys()
  tiles = {}
  for i, ex in enumerate(extents):
    tiles[ex] = ctx.create(tile.from_shape(ex.shape, dtype), hint=i)

  for ex in extents:
    tiles[ex] = tiles[ex].wait().blob_id

  return DistArrayImpl(shape=shape, dtype=dtype, tiles=tiles, reducer_fn=reducer)

def from_table(extents):
  '''
  Construct a distarray from an existing table.
  Keys must be of type `Extent`, values of type `Tile`.
  
  Shape is computed as the maximum range of all extents.
  
  Dtype is taken from the dtype of the tiles.
  
  :param table:
  '''
  Assert.no_duplicates(extents)
  
  if not extents:
    shape = tuple()
  else:
    shape = extent.find_shape(extents.keys())
  
  if len(extents) > 0:
    # fetch a one element array in order to get the dtype
    key, blob_id = extents.iteritems().next()
    #util.log_info('%s :: %s', key, blob_id)
    t = blob_ctx.get().get(blob_id, None)
    Assert.isinstance(t, tile.Tile)
    dtype = t.dtype
  else:
    # empty table; default dtype.
    dtype = np.float

  return DistArrayImpl(shape=shape, dtype=dtype, tiles=extents, reducer_fn=None)

class LocalWrapper(DistArray):
  '''
  Provide the `DistArray` interface for local data.
  '''
  def __init__(self, data):
    self._data = np.asarray(data)
    assert not isinstance(data, core.BlobId)
    #print 'Wrapping: %s %s (%s)' % (data, type(data), np.isscalar(data))
    #print 'DATA: %s' % type(self._data)

  @property
  def dtype(self):
    return self._data.dtype

  @property
  def shape(self):
    return self._data.shape

  def fetch(self, ex):
    return self._data[ex.to_slice()]

  def map_to_array(self, mapper_fn, kw=None):
    if kw is None: kw = {}
    ex = extent.from_slice(np.index_exp[:], self.shape)
    result = mapper_fn(ex, self._data, **kw)
    assert len(result) == 1
    result_ex, result_data = result[0]

    return as_array(result_data)

  def __getitem__(self, idx):
    return self._data[idx]

def as_array(data):
  if isinstance(data, DistArray):
    return data

  # TODO(power) -- promote numpy arrays to distarrays?
  return LocalWrapper(data)


def best_locality(array, ex):
  '''
  Return the table shard with the best locality for extent `ex`.
  :param table:
  :param ex:
  '''
  splits = extent.find_overlapping(array.extents, ex)
  counts = collections.defaultdict(int)
  for key, overlap in splits:
    shard = array.extents[key]
    counts[shard] += overlap.size
  
  s_counts = sorted(counts.items(), key=lambda kv: kv[1])
  return s_counts[-1][0]
  


def _slice_mapper(ex, **kw):
  '''
  Run when mapping over a slice.
  Computes the intersection of the current tile and a global slice.
  If the slice is non-zero, then run the user mapper function.
  Otherwise, do nothing.
  
  :param ex:
  :param tile: 
  :param mapper_fn: User mapper function.
  :param slice: `TileExtent` representing the slice of the input array.
  '''

  mapper_fn = kw['_slice_fn']
  slice_extent = kw['_slice_extent']

  fn_kw = kw['fn_kw']
  if fn_kw is None: fn_kw = {}

  intersection = extent.intersection(slice_extent, ex)
  if intersection is None:
    return []

  offset = extent.offset_from(slice_extent, intersection)
  offset.array_shape = slice_extent.shape

  subslice = extent.offset_slice(ex, intersection)

  result = mapper_fn(offset, **fn_kw)
  #util.log_info('Slice mapper[%s] %s %s -> %s', mapper_fn, offset, subtile, result)
  return result

class Slice(DistArray):
  def __init__(self, darray, idx):
    if not isinstance(idx, extent.TileExtent):
      idx = extent.from_slice(idx, darray.shape)
    util.log_info('New slice: %s', idx)
    
    Assert.isinstance(darray, DistArray)
    self.darray = darray
    self.slice = idx
    self.shape = self.slice.shape
    intersections = [extent.intersection(self.slice, ex) for ex in self.darray.tiles]
    intersections = [ex for ex in intersections if ex is not None]
    offsets = [extent.offset_from(self.slice, ex) for ex in intersections]
    self.tiles = offsets
    self.dtype = darray.dtype
    
  def map_to_table(self, mapper_fn, kw):
    return self.darray.map_to_table(mapper_fn = _slice_mapper,
                                    kw={'fn_kw' : kw,
                                        '_slice_extent' : self.slice,
                                        '_slice_fn' : mapper_fn })

  def fetch(self, idx):
    offset = extent.compute_slice(self.slice, idx.to_slice())
    return self.darray.fetch(offset)



def broadcast_mapper(ex, tile, mapper_fn=None, bcast_obj=None):
  raise NotImplementedError

class Broadcast(DistArray):
  '''A broadcast object mimics the behavior of Numpy broadcasting.
  
  Takes an input of shape (x, y) and a desired output shape (x, y, z),
  the broadcast object reports shape=(x,y,z) and overrides __getitem__
  to return the appropriate values.
  '''
  def __init__(self, base, shape):
    Assert.isinstance(base, (np.ndarray, DistArray))
    Assert.isinstance(shape, tuple)
    self.base = base
    self.shape = shape
  
  def __repr__(self):
    return 'Broadcast(%s -> %s)' % (self.base, self.shape)
  
  def fetch(self, ex):
    # drop extra dimensions
    while len(ex.shape) > len(self.base.shape):
      ex = extent.drop_axis(ex, -1)
      
    # fold down expanded dimensions
    ul = []
    lr = []
    for i in xrange(len(self.base.shape)):
      size = self.base.shape[i]
      if size == 1:
        ul.append(0)
        lr.append(1)
      else:
        ul.append(ex.ul[i])
        lr.append(ex.lr[i])
    
    ex = extent.create(ul, lr, self.base.shape) 

    template = np.ndarray(ex.shape, dtype=self.base.dtype)
    fetched = self.base.fetch(ex)
    
    _, bcast = np.broadcast_arrays(template, fetched)
    return bcast 


def broadcast(args):
  if len(args) == 1:
    return args

  orig_shapes = [list(x.shape) for x in args]
  dims = [len(shape) for shape in orig_shapes]
  max_dim = max(dims)
  new_shapes = []
  
  # prepend filler dimensions for smaller arrays
  for i in range(len(orig_shapes)):
    diff = max_dim - len(orig_shapes[i])
    new_shapes.append([1] * diff + orig_shapes[i])
 
  # check shapes are valid
  # for each axis, all arrays should either share the 
  # same size, or have size == 1
  for axis in range(max_dim):
    axis_shape = set(s[axis] for s in new_shapes)
   
    assert len(axis_shape) <= 2, 'Mismatched shapes for broadcast: %s' % orig_shapes
    if len(axis_shape) == 2:
      assert 1 in axis_shape, 'Mismatched shapes for broadcast: %s' % orig_shapes
  
    # now lift the inputs with size(axis) == 1 
    # to have the maximum size for the axis 
    max_size = max(s[axis] for s in new_shapes)
    for s in new_shapes:
      s[axis] = max_size
    
  # wrap arguments with missing dims in a Broadcast object.
  results = []
  for i in range(len(args)):
    if new_shapes[i] == orig_shapes[i]:
      results.append(args[i])
    else:
      results.append(Broadcast(args[i], tuple(new_shapes[i])))
    
  #util.log_debug('Broadcast result: %s', results)
  return results

def _size(v):
  if isinstance(v, Broadcast):
    return np.prod(v.base.shape)
  return np.prod(v.shape)

def largest_value(vals):
  return sorted(vals, key=lambda v: _size(v))[-1]




#!/usr/bin/env python

"""
Configuration options and flags.

Options may be specified on the command line, or via a configuration
file.  Configuration files should be placed in $HOME/.config/spartan.ini

"""
import ConfigParser
import argparse
import logging
import os
import sys

import appdirs

from spartan import util

class Flag(object):
  def __init__(self, name, default=None, help=''):
    self.name = name
    self.val = default
    self.help = help

  def __repr__(self):
    return '--%s=%s' % (self.name, self._str())

  def _str(self):
    return str(self.val)


class IntFlag(Flag):
  def parse(self, str):
    self.val = int(str)

class StrFlag(Flag):
  def parse(self, str):
    self.val = str

class BoolFlag(Flag):
  def parse(self, str):
    str = str.lower()
    str = str.strip()

    if str == 'false' or str == '0': val = False
    else: val = True
    #print 'Bool %s "%s" %s' % (self.name, str, val)
    self.val = val

  def _str(self):
    return str(int(self.val))


LOG_STR = {logging.DEBUG: 'DEBUG',
           logging.INFO: 'INFO',
           logging.WARN: 'WARN',
           logging.ERROR: 'ERROR',
           logging.FATAL: 'FATAL'}


class LogLevelFlag(Flag):
  def parse(self, str):
    self.val = getattr(logging, str)

  def _str(self):
    return LOG_STR[self.val]



class Flags(object):
  def __init__(self):
    self._parsed = False
    self._vals = {}

  def add(self, flag):
    self._vals[flag.name] = flag

  def __getattr__(self, key):
    if key.startswith('_'): return self.__dict__[key]

    assert self.__dict__['_parsed'], 'Access to flags before config.initialize() called.'
    return self.__dict__['_vals'][key].val

  def __repr__(self):
    return ' '.join([repr(f) for f in self._vals.values()])

  def __str__(self):
    return repr(self)

  def __iter__(self):
    return iter(self._vals.items())


FLAGS = Flags()

FLAGS.add(BoolFlag('profile_kernels', default=False))
FLAGS.add(BoolFlag('profile_master', default=False))
FLAGS.add(BoolFlag('cluster', default=False))
FLAGS.add(LogLevelFlag('log_level', logging.INFO))
FLAGS.add(IntFlag('num_workers', default=3))
FLAGS.add(IntFlag('port_base', default=10000,
                  help='Port to listen on (master = port_base, workers=port_base + N)'))

def initialize(argv):
  '''Parse configuration from flags and/or configuration file.'''

  import spartan.expr.optimize
  import spartan.cluster

  # load flags from other packages (is there a better way to do this?)
  if FLAGS._parsed:
    return

  FLAGS._parsed = True

  config_file = appdirs.user_data_dir('Spartan', 'rjpower.org') + '/spartan.ini'
  config_dir = os.path.dirname(config_file)
  if not os.path.exists(config_dir):
    os.makedirs(config_dir, mode=0755)

  if not os.path.exists(config_file):
    open(config_file, 'a').close()

  print >>sys.stderr, 'Loading configuration from %s' % (config_file)
  try:
    config = ConfigParser.ConfigParser()
    config.read(config_file)

    if config.has_section('flags'):
      for name, value in config.items('flags'):
        argv.append('--%s=%s' % (name, value))
  except:
    print >>sys.stderr, 'Failed to parse config file: %s' % config_file
    sys.exit(1)

  parser = argparse.ArgumentParser()
  for name, flag in FLAGS:
    parser.add_argument('--' + name, type=str)

  parsed_flags, rest = parser.parse_known_args(argv)
  for name, flag in FLAGS:
    if getattr(parsed_flags, name) is not None:
      flag.parse(getattr(parsed_flags, name))

  logging.basicConfig(format='%(filename)s:%(lineno)s [%(funcName)s] %(message)s',
                      level=FLAGS.log_level,
                      stream=sys.stderr)

  for f in rest:
    if f.startswith('-'):
      util.log_warn('Unknown flag: %s (ignored)' % f)

  util.log_debug('Hostlist: %s', FLAGS.hosts)
  return rest




#!/usr/bin/python

from setuptools import setup, Extension
from Cython.Build import cythonize

setup(
  name='spartan',
  version='0.06',
  maintainer='Russell Power',
  maintainer_email='russell.power@gmail.com',
  url='http://github.com/rjpower/spartan',
  install_requires=[
    'appdirs',
    'numpy>=1.6',
    'cython',
    'sphinx_bootstrap_theme',                
  ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Environment :: Other Environment',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: POSIX',
    'Programming Language :: C++',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
  ],
  description='Distributed Numpy-like arrays.',
  package_dir={'': '.'},
  packages=['spartan', 
            'spartan.expr',
            'spartan.rpc',
            'spartan.array' ],
  ext_modules = cythonize(
              'spartan/core.pyx', 
              'spartan/examples/netflix_core.pyx')
)



from .common import *
from .zeromq import *

def listen(host, port):
  socket = server_socket((host, port))
  server = Server(socket)
  return server

def connect(host, port):
  socket = client_socket((host, port))
  return Client(socket)  



import os.path
import socket
import subprocess
import threading
import time

from spartan import config, util
from spartan.config import FLAGS, BoolFlag
import spartan
import spartan.worker
import spartan.master

class HostListFlag(config.Flag):
  def parse(self, str):
    hosts = []
    for host in str.split(','):
      hostname, count = host.split(':')
      hosts.append((hostname, int(count)))
    self.val = hosts

  def _str(self):
    return ','.join(['%s:%d' % (host, count) for host, count in self.val])

class AssignMode(object):
  BY_CORE = 1
  BY_NODE = 2

class AssignModeFlag(config.Flag):
  def parse(self, str): return getattr(AssignMode, str)
  def _str(self):
    if self.val == AssignMode.BY_CORE: return 'BY_CORE'
    return 'BY_NODE'

FLAGS.add(HostListFlag('hosts', default=[('localhost', 8)]))
FLAGS.add(BoolFlag('xterm', default=False, help='Run workers in xterm'))
FLAGS.add(BoolFlag('oprofile', default=False, help='Run workers inside of operf'))
FLAGS.add(AssignModeFlag('assign_mode', default=AssignMode.BY_CORE))

FLAGS.add(BoolFlag(
  'use_threads',
  help='When running locally, use threads instead of forking. (slow, for debugging)',
  default=True))


def _start_remote_worker(worker, st, ed):
  if FLAGS.use_threads and worker == 'localhost':
    util.log_info('Using threads.')
    for i in range(st, ed):
      p = threading.Thread(target=spartan.worker._start_worker,
                           args=((socket.gethostname(), FLAGS.port_base), FLAGS.port_base + 1 + i, i))
      p.daemon = True
      p.start()
    time.sleep(0.1)
    return

  util.log_info('Starting worker %d:%d on host %s', st, ed, worker)
  if FLAGS.oprofile:
    os.system('mkdir operf.%s' % worker)
    
  ssh_args = ['ssh', '-oForwardX11=no', worker ]
 
  args = ['cd %s && ' % os.path.abspath(os.path.curdir)]

  if FLAGS.xterm:
    args += ['xterm', '-e',]

  if FLAGS.oprofile:
    args += ['operf -e CPU_CLK_UNHALTED:100000000', '-g', '-d', 'operf.%s' % worker]

  args += [
          #'gdb', '-ex', 'run', '--args',
          'python', '-m spartan.worker',
          '--master=%s:%d' % (socket.gethostname(), FLAGS.port_base),
          '--count=%d' % (ed - st),
          '--port=%d' % (FLAGS.port_base + 1)]

  # add flags from config/user
  args += repr(FLAGS).split(' ')

  util.log_debug('Running worker %s', ' '.join(args))
  time.sleep(0.1)
  if worker != 'localhost':
    p = subprocess.Popen(ssh_args + args, executable='ssh')
  else:
    p = subprocess.Popen(' '.join(args), shell=True, stdin=subprocess.PIPE)
    
  return p

def start_cluster(num_workers, use_cluster_workers):
  '''
  Start a cluster with ``num_workers`` workers.
  
  If use_cluster_workers is True, then use the remote workers
  defined in `spartan.config`.  Otherwise, workers are all
  spawned on the localhost.
  
  :param num_workers:
  :param use_cluster_workers:
  '''
  if not use_cluster_workers:
    _start_remote_worker('localhost', 0, num_workers)
  else:
    available_workers = sum([cnt for _, cnt in FLAGS.hosts])
    assert available_workers > num_workers, 'Insufficient slots to run all workers.'
    count = 0
    num_hosts = len(FLAGS.hosts)
    for worker, total_tasks in FLAGS.hosts:
      if FLAGS.assign_mode == AssignMode.BY_CORE:
        sz = total_tasks
      else:
        sz = util.divup(num_workers, num_hosts)
      
      sz = min(sz, num_workers - count)
      _start_remote_worker(worker, count, count + sz)
      count += sz
      if count == num_workers:
        break

  master = spartan.master.Master(FLAGS.port_base, num_workers)
  time.sleep(0.1)
  master.wait_for_initialization()
  return master




#!/usr/bin/env python

"""
Lazy expression DAGs.

All expression operations are subclasses of the `Expr` class, which
by default performs all operations lazily.

Operations are built up using a few high-level operations -- these all
live in their own modules:

* Create a new distributed array `spartan.expr.ndarray`
* Map over an array :py:mod:`spartan.expr.map` and `spartan.expr.shuffle`
* Reduce over an array `spartan.expr.reduce`
* Apply a stencil/convolution to an array `spartan.expr.stencil`
* Slicing/indexing `spartan.expr.index`.   

Optimizations on DAGs live in `spartan.expr.optimize`.

"""

from base import Expr, evaluate, dag, glom, eager, lazify, force, make_primitive, \
  NotShapeable
import numpy as np
from spartan import util

from ..array import extent
from ..array.extent import index_for_reduction, shapes_match
from ..util import Assert
from .builtins import *
from .map import map
from .ndarray import ndarray
from .outer import outer
from .reduce import reduce
from .shuffle import shuffle


Expr.outer = outer
Expr.sum = sum
Expr.mean = mean
Expr.astype = astype
Expr.ravel = ravel
Expr.argmin = argmin
Expr.argmax = argmax



import math

from ..node import Node
from spartan import util
from spartan.array import extent
import numpy as np
from spartan.util import divup, Assert


try:
  import parakeet
  jit = parakeet.jit
except:
  def jit(fn):
    return fn
  
def tiles_like(array, target_shape):
  orig_shape = array.shape
  orig_tile = array.tile_shape()
  new_tile = []
  for i in range(len(orig_shape)):
    scale = float(target_shape[i]) / orig_shape[i]    
    new_tile.append(int(math.ceil(orig_tile[i] * scale)))
  return new_tile

@util.synchronized
def convolve(local_image, local_filters):
  num_images, n_color, w, h = local_image.shape
  num_filts, f_col, fw, fh = local_filters.shape

  def _inner(args):
    iid, fid, x, y = args
    image = local_image[iid]
    f = local_filters[fid]
    out = 0
    for c in xrange(n_color):
      for i in xrange(fw):
        for j in xrange(fh):
          if x + i < w and y + j < h:
            out += image[c, x + i, y + j] * f[c, i, j]
    return out

  return parakeet.imap(_inner, (num_images, num_filts, w, h))

def _divup(a, b):
  return int(math.ceil(float(a) / float(b)))

@util.synchronized
@jit
def _maxpool(array, pool_size, stride):
  n, c, w, h = array.shape
  target = np.ones((n, c,
                    _divup(w, stride),
                    _divup(h, stride))) * -1e12
  
  for img in xrange(n):
    for color in xrange(c):
      for x in xrange(0, w, stride): 
        for y in xrange(0, h, stride): 
          for i in xrange(0, pool_size):
            for j in xrange(0, pool_size):
              if x + i < w and y + j < h:
                target[0, 0, 0, 0] = 0
                # target[img, color, (x + i) / stride, (y + j) / stride] = \
                #  max(array[img, color, x + i, y + j],
                #      target[img, color, x + i, y + j])
  
  return target
                      
def stencil_mapper(array, ex, filters=None, images=None, target_shape=None):
  local_filters = filters.glom()
  # util.log_info('R:%s', region)
  # util.log_info('F:%s', local_filters.shape)
  local_image = images.fetch(ex)
  
  num_img, n_col, w, h = images.shape
  num_filt, f_col, fw, fh = filters.shape
  
  Assert.eq(n_col, f_col)
  
  util.log_info('Stencil(%s), image: %s, filter: %s (%s, %s)',
            ex,
            local_image.shape, local_filters.shape,
            images.shape, filters.shape)
  
  target_ex = extent.create(
      (ex.ul[0], 0, ex.ul[2], ex.ul[3]),
      (ex.lr[0], num_filt, ex.lr[2], ex.lr[3]),
      target_shape)

  result = convolve(local_image, local_filters)
  
  util.log_info('Updating: %s', target_ex)
  yield (target_ex, result)


def stencil(images, filters, stride=1):
  from .base import eager
  from .ndarray import ndarray
  from .shuffle import shuffle
  images = eager(images)
  filters = eager(filters)
  
  images = images.force()
 
  n_img, n_col, w, h = images.shape
  n_filt, f_col, fw, fh = filters.shape
  
  tile_hint = tiles_like(images, (n_img, n_filt, w, h))
#     util.log_info('Stencil: %s %s %s',
#                   images.shape, (n_img, n_filt, w, h),
#                   tile_hint)
   
  target = ndarray((n_img, n_filt, w, h),
                   dtype=images.dtype,
                   reduce_fn=np.sum,
                   combine_fn=np.sum,
                   tile_hint=tile_hint)
  
  return shuffle(images,
                     stencil_mapper,
                     target=target,
                     kw=dict(images=images,
                             filters=filters,
                             target_shape=target.shape))


def _maxpool_mapper(array, ex, pool_size, stride, target_shape):
  region = array.fetch(ex)
  # util.log_info('%s %s', inputs[0].shape, region.shape)
  pooled = _maxpool(region, pool_size, stride)
  ul = ex.ul
  lr = ex.lr
  sz = ex.shape
  
  t_ul = ul[:2] + tuple(np.array(ul[2:]) / stride)
  t_lr = lr[:2] + divup(lr[2:], stride)
  
  target_ex = extent.create(t_ul, t_lr, target_shape)
  
  # util.log_info('%s %s %s', ex, target_ex, pooled.shape)
  
  yield (target_ex, pooled)
  
def maxpool(images, pool_size=2, stride=2):
  from .shuffle import shuffle
  from .ndarray import ndarray
  
  images = images.force()
  n_img, n_col = images.shape[:2]
  tgt_shape = divup(images.shape[2:], stride)
  tile_hint = tiles_like(images, (n_img, n_col,) + tgt_shape)
  
  util.log_info('%s %s %s %s',
                images.shape[2:], tgt_shape, images.tile_shape(), tile_hint)
  target = ndarray((n_img, n_col) + tgt_shape,
                   dtype=images.dtype,
                   tile_hint=tile_hint,
                   combine_fn=np.maximum,
                   reduce_fn=np.maximum)
   
  return shuffle(images,
                    _maxpool_mapper,
                    target=target,
                    kw=dict(target_shape=target.shape,
                            stride=stride,
                            pool_size=pool_size))
        


    



#!/usr/bin/env python

import cProfile
import pstats
import threading
import time

try:
  import pyximport; pyximport.install()
except ImportError:
  pass

import multiprocessing
from multiprocessing.pool import ThreadPool
import os
import socket
import sys

import resource
from spartan import config, util, rpc, core, blob_ctx
from spartan.config import FLAGS, StrFlag, IntFlag

class Worker(object):
  def __init__(self, port, master):
    self.id = -1
    self._initialized = False
    self._peers = {}
    self._blobs = {}
    self._master = master
    self._running = True
    self._ctx = None
    self._kernel_threads = ThreadPool(processes=1)
    self._kernel_prof = cProfile.Profile()

    hostname = socket.gethostname()
    self._server = rpc.listen(hostname, port)
    self._server.register_object(self)
    self._server.serve_nonblock()

    req = core.RegisterReq()
    req.host = hostname
    req.port = port
    master.register(req)


  def get_worker(self, worker_id):
    return self._peers[worker_id]

  def initialize(self, req, handle):
    util.log_info('Worker %d initializing...', req.id)

    for id, (host, port) in req.peers.iteritems():
      self._peers[id] = rpc.connect(host, port)

    self.id = req.id
    self._ctx = blob_ctx.BlobCtx(self.id, self._peers, self)
    self._initialized = True
    handle.done()
    
  def create(self, req, handle):
    assert self._initialized
    blob_ctx.set(self._ctx)

    if req.blob_id.id == -1:
      id = blob_ctx.get().create_local()
    else:
      id = req.blob_id

    self._blobs[id] = req.data
    resp = core.CreateResp(blob_id=id)
    #util.log_info('W%d :: Created blob: id %s', self.id, id)
    handle.done(resp)

  def destroy(self, req, handle):
    for id in req.ids:
      if id in self._blobs:
        del self._blobs[id]

    #util.log_info('Destroy...')
    handle.done()

  def update(self, req, handle):
    blob =  self._blobs[req.id]
    self._blobs[req.id] = blob.update(req.data, req.reducer)

    handle.done()

  def get(self, req, handle):
    if req.selector is None:
      #util.log_info('GET: %s', type(self._blobs[req.id]))
      resp = core.GetResp(data=self._blobs[req.id])
      handle.done(resp)
    else:
      resp = core.GetResp(data=self._blobs[req.id].get(req.selector))
      handle.done(resp)

  def _run_kernel(self, req, handle):
    self._server._timers['run_kernel'].start()
    if FLAGS.profile_kernels:
      self._kernel_prof.enable()

    try:
      blob_ctx.set(self._ctx)
      results = {}
      for blob_id in req.blobs:
        #util.log_info('%s %s', blob_id, blob_id in self._blobs)
        if blob_id in self._blobs:
          #util.log_info('W%d kernel start', self.id)
          blob = self._blobs[blob_id]
          results[blob_id] = req.mapper_fn(blob_id, blob, **req.kw)
          #util.log_info('W%d kernel finish', self.id)
      handle.done(results)
    except:
      handle.exception()

    self._server._timers['run_kernel'].stop()
    self._kernel_prof.disable()

  def run_kernel(self, req, handle):
    self._kernel_threads.apply_async(self._run_kernel, args=(req, handle))

  def shutdown(self, req, handle):
    if FLAGS.profile_kernels:
      os.system('mkdir -p ./_kernel-profiles/')
      stats = pstats.Stats(self._kernel_prof)
      stats.add(rpc.poller().profiler)
      stats.dump_stats('./_kernel-profiles/%d' % self.id)

    util.log_info('Shutdown worker %d (profile? %d)', self.id, FLAGS.profile_kernels)
    #print self._server.timings()

    handle.done()
    threading.Thread(target=self._shutdown).start()

  def _shutdown(self):
    time.sleep(0.1)
    util.log_info('Closing server...')
    self._running = False
    self._server.shutdown()

  def wait_for_shutdown(self):
    while self._running:
      time.sleep(0.1)


def _start_worker(master, port, local_id):
  util.log_info('Master: %s', master)
  pid = os.getpid()
  os.system('taskset -pc %d %d > /dev/null' % (local_id, pid))
  master = rpc.connect(*master)
  worker = Worker(port, master)
  worker.wait_for_shutdown()

if __name__ == '__main__':
  sys.path.append('./tests')

  FLAGS.add(StrFlag('master'))
  FLAGS.add(IntFlag('port'))
  FLAGS.add(IntFlag('count'))

  resource.setrlimit(resource.RLIMIT_AS, (8 * 1000 * 1000 * 1000,
                                          8 * 1000 * 1000 * 1000))

  config.initialize(sys.argv)
  assert FLAGS.master is not None
  assert FLAGS.port > 0
  assert FLAGS.count > 0

  util.log_info('Starting %d workers on %s', FLAGS.count, socket.gethostname())

  m_host, m_port = FLAGS.master.split(':')
  master = (m_host, int(m_port))

  workers = []
  for i in range(FLAGS.count):
    p = multiprocessing.Process(target=_start_worker, 
                                args=(master, FLAGS.port + i, i))
    p.start()
    workers.append(p)
    
    
  def kill_workers():
    for p in workers:
      p.terminate()
      
  watchdog = util.FileWatchdog(on_closed=kill_workers)
  watchdog.start()
  watchdog.join()



'''
Basic numpy style operations on arrays.

These include --

* Array creation routines: (`rand`, `randn`, `zeros`, `ones`, `arange`)
* Reductions: (`sum`, `argmin`, `argmax`, `mean`)
* Shape/type casting: (`reshape`, `ravel`, `astype`, `shape`, `size`)
* Other: (`dot`).
'''
import numpy as np
from ..array import distarray, extent
from ..array.extent import index_for_reduction, shapes_match
from .base import force
from .shuffle import shuffle
from .map import map
from .ndarray import ndarray
from .reduce import reduce
from .loop import loop
from spartan import util
from ..util import Assert

def rand(*shape, **kw):
  '''
  :param tile_hint: A tuple indicating the desired tile shape for this array.
  '''
  for s in shape: assert isinstance(s, int)
  return map(ndarray(shape, 
                     dtype=np.float,
                     tile_hint=kw.get('tile_hint', None)),
                     fn = lambda input: np.random.rand(*input.shape))
  
def randn(*shape, **kw):
  for s in shape: assert isinstance(s, int)
  return map(ndarray(shape,
                     dtype=np.float,
                     tile_hint=kw.get('tile_hint', None)),
             fn = lambda input: np.random.randn(*input.shape))

def zeros(shape, dtype=np.float, tile_hint=None):
  return map(ndarray(shape, dtype=np.float, tile_hint=tile_hint),
             fn = lambda input: np.zeros(input.shape))

def ones(shape, dtype=np.float, tile_hint=None):
  return map(ndarray(shape, dtype=np.float, tile_hint=tile_hint), 
             fn = lambda input: np.ones(input.shape, dtype))

def _arange_mapper(inputs, ex, dtype=None):
  pos = extent.ravelled_pos(ex.ul, ex.array_shape)
  #util.log_info('Extent: %s, pos: %s', ex, pos)
  sz = np.prod(ex.shape)
  yield (ex, np.arange(pos, pos+sz, dtype=dtype).reshape(ex.shape))


def arange(shape, dtype=np.float):
  return shuffle(ndarray(shape, dtype=dtype), 
                     fn = _arange_mapper, 
                     kw = {'dtype' : dtype })

def _sum_local(ex, tile, axis):
  return np.sum(tile[:], axis)

def sum(x, axis=None):
  '''
  Sum ``x`` over ``axis``.
  
  
  :param x: The array to sum.
  :param axis: Either an integer or ``None``.
  '''
  return reduce(x, axis=axis,
                       dtype_fn = lambda input: input.dtype,
                       local_reduce_fn = _sum_local,
                       combine_fn = lambda a, b: a + b)
    
def mean(x, axis=None):
  '''
  Compute the mean of ``x`` over ``axis``.
  
  See `numpy.ndarray.mean`.
  
  :param x: `Expr`
  :param axis: integer or ``None``
  '''
  return sum(x, axis) / x.shape[axis]


def _to_structured_array(*vals):
  '''Create a structured array from the given input arrays.
  
  :param vals: A list of (field_name, `np.ndarray`)
  :rtype: A structured array with fields from ``kw``.
  '''
  out = np.ndarray(vals[0][1].shape, 
                  dtype=','.join([a.dtype.str for name, a in vals]))
  out.dtype.names = [name for name, a in vals]
  for k, v in vals:
    out[k] = v
  return out

def _take_idx_mapper(input):
  return input['idx']


def _dual_reducer(ex, tile, axis, idx_f=None, val_f=None):
  local_idx = idx_f(tile[:], axis)
  local_val = val_f(tile[:], axis)

  global_idx = ex.to_global(local_idx, axis)
  new_idx = index_for_reduction(ex, axis)
  new_val = _to_structured_array(('idx', global_idx), ('val', local_val))
  
  assert shapes_match(new_idx, new_val), (new_idx, new_val.shape)
  return new_val

def _dual_combiner(a, b, op):
  return np.where(op(a['val'], b['val']), a, b)

def _dual_dtype(input):
  dtype = np.dtype('i8,%s' % input.dtype.str)
  dtype.names = ('idx', 'val')
  return dtype 


def argmin(x, axis=None):
  '''
  Compute argmin over ``axis``.
  
  See `numpy.ndarray.argmin`.
  
  :param x: `Expr` to compute a minimum over. 
  :param axis: Axis (integer or None).
  '''
  compute_min = reduce(x, axis,
                       dtype_fn = _dual_dtype,
                       local_reduce_fn = _dual_reducer,
                       combine_fn = lambda a, b: _dual_combiner(a, b, np.less),
                       fn_kw={'idx_f' : np.argmin, 'val_f' : np.min})
  
  take_indices = map(compute_min, _take_idx_mapper)
  return take_indices

def argmax(x, axis=None):
  '''
  Compute argmax over ``axis``.
  
  See `numpy.ndarray.argmax`.
  
  :param x: `Expr` to compute a maximum over. 
  :param axis: Axis (integer or None).
  '''
  compute_max = reduce(x, axis,
                       dtype_fn = _dual_dtype,
                       local_reduce_fn = _dual_reducer,
                       combine_fn = lambda a, b: _dual_combiner(a, b, np.greater),
                       fn_kw={'idx_f' : np.argmax, 'val_f' : np.max})
  
  take_indices = map(compute_max, _take_idx_mapper)
  return take_indices
  

def size(x, axis=None):
  '''
  Return the size (product of the size of all axes) of ``x``.
  
  See `numpy.ndarray.size`.
  
  :param x: `Expr` to compute the size of.
  '''
  if axis is None:
    return np.prod(x.shape)
  return x.shape[axis]

def astype(x, dtype):
  '''
  Convert ``x`` to a new dtype.
  
  See `numpy.ndarray.astype`.
  
  :param x:
  :param dtype:
  '''
  assert x is not None
  return map(x, lambda tile: tile.astype(dtype))

def _ravel_mapper(array, ex):
  ul = extent.ravelled_pos(ex.ul, ex.array_shape)
  lr = 1 + extent.ravelled_pos([lr - 1 for lr in ex.lr], ex.array_shape)
  shape = (np.prod(ex.array_shape),)
  
  ravelled_ex = extent.create((ul,), (lr,), shape)
  ravelled_data = array.fetch(ex).ravel()
  yield ravelled_ex, ravelled_data
   
def ravel(v):
  '''
  "Ravel" ``v`` to a one-dimensional array of shape (size(v),).
  
  See `numpy.ndarray.ravel`.
  :param v:
  '''
  return shuffle(v, _ravel_mapper)

def _reshape_mapper(array, ex, _dest_shape):
  tile = array.fetch(ex)
  
  ravelled_ul = extent.ravelled_pos(ex.ul, ex.array_shape)
  ravelled_lr = extent.ravelled_pos([lr - 1 for lr in ex.lr], ex.array_shape)
  
  target_ul = extent.unravelled_pos(ravelled_ul, _dest_shape)
  target_lr = extent.unravelled_pos(ravelled_lr, _dest_shape)
  
  #util.log_info('%s + %s -> %s', ravelled_ul, _dest_shape, target_ul)
  #util.log_info('%s + %s -> %s', ravelled_lr, _dest_shape, target_lr)
  
  target_ex = extent.create(target_ul, np.array(target_lr) + 1, _dest_shape)
  yield target_ex, tile.reshape(target_ex.shape)

def reshape(array, new_shape, tile_hint=None):
  '''
  Reshape/retile ``array``.
  
  Returns a new array with the given shape and tile size.
  
  :param array: `Expr`
  :param new_shape: `tuple`
  :param tile_hint: `tuple` or None
  '''
  
  old_size = np.prod(array.shape)
  new_size = np.prod(new_shape)
  
  Assert.eq(old_size, new_size, 'Size mismatch')
  
  return shuffle(array, 
                     _reshape_mapper, 
                     tile_hint=tile_hint,
                     kw = { '_dest_shape' : new_shape})

def _dot_mapper(inputs, ex, av, bv):
  # read current tile of array 'a'
  ex_a = ex

  # fetch corresponding column tile of array 'b'
  # rows = ex_a.cols
  # cols = *
  ex_b = extent.create((ex_a.ul[1], 0),
                       (ex_a.lr[1], bv.shape[1]),
                       bv.shape)

  util.log_info('%s %s', ex_a, ex_b)

  a = av.fetch(ex_a)
  b = bv.fetch(ex_b)

  result = np.dot(a, b)

  ul = np.asarray([ex_a.ul[0], 0])
  lr = ul + result.shape
  #util.log_info('%s %s %s', a.shape, b.shape, result.shape)
  #util.log_info('%s %s %s', ul, lr, target_shape)
  target_shape = (av.shape[0], bv.shape[1])
  out = extent.create(ul, lr, target_shape)
  
  yield out, result

def _dot_numpy(array, ex, numpy_data=None):
  l = array.fetch(ex)
  r = numpy_data
  
  yield (ex[0].add_dim(), np.dot(l, r))
  

def dot(a, b):
  '''
  Compute the dot product (matrix multiplication) of 2 arrays.
  
  :param a: `Expr` or `numpy.ndarray` 
  :param b: `Expr` or `numpy.ndarray`
  :rtype: `Expr`
  '''
  av = force(a)
  bv = force(b)
  
  if isinstance(bv, np.ndarray):
    return shuffle(av, _dot_numpy, kw = { 'numpy_data' : bv })
  
  #av, bv = distarray.broadcast([av, bv])
  Assert.eq(a.shape[1], b.shape[0])
  target = ndarray((a.shape[0], b.shape[1]),
                   dtype=av.dtype,
                   tile_hint=av.tile_shape(),
                   combine_fn=np.add,
                   reduce_fn=np.add)
  
  return shuffle(av, _dot_mapper, target=target, kw = dict(av=av, bv=bv))
            

def ln(v): return map(v, fn=np.log)
def log(v): return map(v, fn=np.log)
def exp(v): return map(v, fn=np.exp)
def sqrt(v): return map(v, fn=np.sqrt)
def abs(v): return map(v, fn=np.abs)

try:
  import scipy.stats
  def norm_cdf(v):
    return map(v, fn=scipy.stats.norm.cdf)
except:
  util.log_info('Missing scipy.stats (some functions will be unavailable.')
